#include "pool/test.h"
#include "pool/TestBase.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/DbSelect.h"
#include "StorageSvc/DbOption.h"
#include "StorageSvc/DbColumn.h"
#include "StorageSvc/DbTypeInfo.h"
#include "StorageSvc/IStorageSvc.h"
#include "StorageSvc/IStorageExplorer.h"
#include "StorageSvc/FileDescriptor.h"

namespace pool  {

  /** @class ObjectIterator
    *
    */
  class ObjectIterator : public TestBase {
  public: 

    /// Standard Constructor
    ObjectIterator() : TestBase(ROOTKEY_StorageType.type(), "test_1_1.root", "test_1_2.root") {
      m_criteria = "m_char<=3";
    }

    /// Standard Destructor
    virtual ~ObjectIterator()               {    }

    /// Read objects from the persistent store
    DbStatus read()  {
      SessionH sesH = 0;
      DbStatus sc = DbStatus::Error;
      std::string criteria = m_criteria;
      IStorageSvc* storSvc = storageSvc();
      IStorageExplorer* exp = 0;
      sc = storSvc->queryInterface(IStorageExplorer::interfaceID(),(void**)&exp);
      checkError(sc,"queryInterface(IStorageExplorer::interfaceID(),(void**)&exp)");
      if ( !exp ) {
        std::cout << "Cannot access storage explorer object." << std::endl;
        return DbStatus::Error;
      }
      sc = exp->startSession(poolDb::READ, m_type, sesH);
      checkError(sc,"startSession(poolDb::READ, m_type, sesH)");

      FileDescriptor dsc2(m_dbName2, m_dbName2);
      sc = exp->connect(sesH, poolDb::READ, dsc2);
      checkError(sc,"connect(sesH, poolDb::READ, dsc2)");

      std::vector<const Token*> conts;
      sc = exp->containers(dsc2, conts);
      std::cout << " ============== Execute SELECT ==========" << std::endl;
      for(std::vector<const Token*>::iterator k1=conts.begin(); k1 != conts.end(); ++k1) {
        if ( (*k1)->contID() == "test_cont_2" )  {
          std::cout << " ============== SELECT on Tree:" 
                    << (*k1)->contID() << " ==========" << std::endl;
          Token* tok = 0;
          DbSelect sel(criteria);
          std::string cntName = (*k1)->contID();
          sc = exp->select(dsc2, cntName, sel);
          checkError(sc,"select(dsc2, cntName, sel)");
          while ( sc.isSuccess() )  {
            sc = exp->next(sel,tok);
            checkError(sc,"next(sel,tok)");
            if ( sc.isSuccess() )  {
              std::cout << tok->toString() << std::endl;
              tok->release();
            }
          }
        }
      }


      FileDescriptor dsc1(m_dbName1, m_dbName1);
      sc = exp->connect(sesH, poolDb::READ, dsc1);
      checkError(sc,"connect(sesH, poolDb::READ, dsc1)");

      sc = exp->containers(dsc1, conts);
      for(std::vector<const Token*>::iterator k2=conts.begin(); k2 != conts.end(); ++k2)  {
        if ( (*k2)->contID() == "test_cont_1" )  {
          std::cout << " ============== SELECT on Directory:" 
                    << (*k2)->contID() << " ==========" << std::endl;
          Token* tok1 = 0, *tok2 = 0;
          cache()->setPrintout(1);
          DbSelect sel1("");
          std::string cntName = (*k2)->contID();
          sc = exp->select(dsc1, cntName, sel1);
          checkError(sc,"select(dsc1, cntName, sel1)");

          DbSelect sel2("");
          sc = exp->select(dsc1, cntName, sel2);
          checkError(sc,"select(dsc1, cntName, sel2)");

          size_t num_token = 0, num_object = 0;
          while ( sc.isSuccess() )  {
            try {
              sc = exp->next(sel1,tok1);
              checkError(sc,"next(sel,tok)");
              sc = exp->next(sel2,tok2);
              checkError(sc,"next(sel2,tok2)");
              Ref<TNamed> ref2(cache(), tok2);
              Ref<TNamed> ref1(cache(), tok1);
              if ( ref2 )  {
                const char* name = ref2->GetTitle();
                if ( name )  {
                  num_object++;
                }
              }
              if ( ref1 )  {
                const char* name = ref1->GetTitle();
                if ( name )  {
                  num_object++;
                }
              }

              if ( ref1.ptr() && ref2.ptr() )  {
                if ( ref1.ptr() != ref2.ptr() )  {
                  std::cout << "Different pointers!" << std::endl;
                }
                else {
                  if ( ref1.token()->toString() != ref2.token()->toString() )  {
                    std::cout << "Different tokens!" << std::endl;
                  }
                  if ( ++num_token < 20 )  {
                    std::cout << "Ref1 token:" << tok1->toString() << std::endl
                              << "    Type: " << typeid(*(ref1.ptr())).name()
                              << std::endl;
                  }
                  else {
                    cache()->setPrintout(0);
                  }
                }
              }
            }
            catch(std::exception& e)  {
              std::cout << "End of token loop:" << e.what() << std::endl;
            }
          }
          std::cout << "Successfully received " << num_token  << " tokens." << std::endl;
          std::cout << "Successfully received " << num_object << " objects." << std::endl;
          cache()->clear();
        }
      }

      std::cout << " ========================================" << std::endl;
      cache()->disconnect(0);
      std::cout << "DbStorageSvc::release()=" << exp->release() << std::endl;
      return sc;
    }

    /// Read objects from the persistent store
    DbStatus write()  {
      DbStatus sc = DbStatus::Error;
      std::string criteria = m_criteria;
      IStorageSvc* storSvc = storageSvc();
      IStorageExplorer* exp = 0;
      sc = storSvc->queryInterface(IStorageExplorer::interfaceID(),(void**)&exp);
      checkError(sc,"queryInterface(IStorageExplorer::interfaceID(),(void**)&exp)");
      if ( !exp ) {
        std::cout << "Cannot access storage explorer object." << std::endl;
        return DbStatus::Error;
      }
      ConnectionH dbH = 0;
      sc = cache()->connect(m_dbName1, poolDb::RECREATE, dbH);
      checkError(sc,"connect(m_dbName1, poolDb::RECREATE, dbH)");
      FileDescriptor dsc1(m_dbName1, m_dbName1);
      dsc1.setDbc(dbH);

      Guid g3(getGuid("AnotherTNamed"));
      Guid g4(getGuid("A::B::C::Named"));

      Placement place1(m_dbName1,"test_cont_1", getGuid("TNamed"), ROOTKEY_StorageType.type());
      Placement place3(m_dbName1,"test_cont_1", g3, ROOTKEY_StorageType.type());
      Placement place4(m_dbName1,"test_cont_1", g4, ROOTKEY_StorageType.type());

      TransactionH trH1 = 0;

      for ( size_t j = 0; j < m_outerLoops; ++j )   {
        sc = cache()->startTransaction(dsc1.dbc(), trH1);
        checkError(sc,"Cannot start database transactions (1)");
        for ( size_t i = 0; i < m_innerLoops; ++i )   {
          char txt1[256], txt2[256], txt3[256], txt4[256], txt5[256];
          ::sprintf(txt1,"Hello [TNamed(%d)]", int(i));
          ::sprintf(txt2,"World [TNamed(%d)]", int(i));
          ::sprintf(txt3,"Hello [test(%d)]", int(i));
          ::sprintf(txt4,"AnotherTNamed_%d", int(i));
          ::sprintf(txt5,"A::B::C::Named_%d", int(i));

          pool::Ref<A::B::C::Named> refNS(new A::B::C::Named(txt5, txt3));
          pool::Ref<AnotherTNamed> refAnother(new AnotherTNamed(txt4, txt3));
          pool::Ref<TNamed> refNamed1(new TNamed(txt1, txt2));

          refAnother->str = "For the sake of it!";
          refNS->str      = "Namespaces are fun!";


          sc = cache()->registerForWrite(trH1, place1, refNamed1.ptr());
          if ( !sc.isSuccess() )   {
            checkError(sc,"Error cache.registerForWrite(trH1, place1, TNamed) [1]");
          }
          sc = cache()->registerForWrite(trH1, place3, refAnother.ptr());
          if ( !sc.isSuccess() )   {
            checkError(sc,"Error cache.registerForWrite(trH1, place3, AnotherTNamed)   [3]");
          }
          sc = cache()->registerForWrite(trH1, place4, refNS.ptr());
          if ( !sc.isSuccess() )   {
            checkError(sc,"Error cache.registerForWrite(trH1, place4, A::B::C::Named) [4]");
          }
        }
        sc = cache()->commit(trH1);
        checkError(sc,"Cannot commit database transaction.");
      }

      sc = cache()->startTransaction(dsc1.dbc(), trH1);
      checkError(sc,"Cannot start database transactions (2)");
      std::vector<const Token*> conts;
      sc = exp->containers(dsc1, conts);
      for(std::vector<const Token*>::iterator k2=conts.begin(); k2 != conts.end(); ++k2)  {
        if ( (*k2)->contID() == "test_cont_1" )  {
          std::cout << " ============== SELECT on Directory:" 
                    << (*k2)->contID() << " ==========" << std::endl;
          Token* tok = 0;
          DbSelect sel("");
          std::string cntName = (*k2)->contID();
          cache()->setPrintout(1);
          sc = exp->select(dsc1, cntName, sel);
          checkError(sc,"select(dsc1, cntName, sel)");

          size_t num_token = 0, num_object = 0;
          while ( sc.isSuccess() )  {
            sc = exp->next(sel,tok);
            checkError(sc,"next(sel,tok)");
            if ( sc.isSuccess() )  {
              Ref<TNamed> ref(cache(), tok);
              if ( ref )  {
                const char* name = ref->GetTitle();
                if ( name )  {
                  num_object++;
                }
              }
              ++num_token;
              if ( num_token < 20 || (num_token > 1200 && num_token < 1300) )  {
                cache()->setPrintout(1);
                std::cout << tok->toString() << std::endl
                          << "    Type: " << typeid(*(ref.ptr())).name()
                          << std::endl;
              }
              else {
                cache()->setPrintout(0);
              }
            }
            else  {
              std::cout << "End of token loop!" << std::endl;
            }
          }
          std::cout << "Successfully received " << num_token  << " tokens." << std::endl;
          std::cout << "Successfully received " << num_object << " objects." << std::endl;
          cache()->clear();
        }
      }
      std::cout << "Rollback transaction..." << std::endl;
      sc = cache()->rollback(trH1);
      checkError(sc,"Error cache.commit(...) [1]");

      std::cout << " ========================================" << std::endl;
      cache()->disconnect(0);
      return sc;
    }

    /// Read objects from the persistent store
    DbStatus update()  {
      DbStatus sc = DbStatus::Error;
      std::string criteria = m_criteria;
      IStorageSvc* storSvc = storageSvc();
      IStorageExplorer* exp = 0;
      sc = storSvc->queryInterface(IStorageExplorer::interfaceID(),(void**)&exp);
      checkError(sc,"queryInterface(IStorageExplorer::interfaceID(),(void**)&exp)");
      if ( !exp ) {
        std::cout << "Cannot access storage explorer object." << std::endl;
        return DbStatus::Error;
      }
      ConnectionH dbH = 0;
      sc = cache()->connect(m_dbName1, poolDb::UPDATE, dbH);
      checkError(sc,"connect(m_dbName1, poolDb::UPDATE, dbH)");
      FileDescriptor dsc1(m_dbName1, m_dbName1);
      dsc1.setDbc(dbH);

      Guid g3(getGuid("AnotherTNamed"));
      Guid g4(getGuid("A::B::C::Named"));

      Placement place1(m_dbName1,"test_cont_1", getGuid("TNamed"), ROOTKEY_StorageType.type());
      Placement place3(m_dbName1,"test_cont_1", g3, ROOTKEY_StorageType.type());
      Placement place4(m_dbName1,"test_cont_1", g4, ROOTKEY_StorageType.type());

      TransactionH trH1 = 0;

      for ( size_t j = 0; j < m_outerLoops; ++j )   {
        sc = cache()->startTransaction(dsc1.dbc(), trH1);
        checkError(sc,"Cannot start database transactions (1)");
        for ( size_t i = 0; i < m_innerLoops; ++i )   {
          char txt1[256], txt2[256], txt3[256], txt4[256], txt5[256];
          ::sprintf(txt1,"Hello [TNamed(%d)]", int(i));
          ::sprintf(txt2,"World [TNamed(%d)]", int(i));
          ::sprintf(txt3,"Hello [test(%d)]", int(i));
          ::sprintf(txt4,"AnotherTNamed_%d", int(i));
          ::sprintf(txt5,"A::B::C::Named_%d", int(i));

          pool::Ref<A::B::C::Named> refNS(new A::B::C::Named(txt5, txt3));
          pool::Ref<AnotherTNamed> refAnother(new AnotherTNamed(txt4, txt3));
          pool::Ref<TNamed> refNamed1(new TNamed(txt1, txt2));

          refAnother->str = "For the sake of it!";
          refNS->str      = "Namespaces are fun!";


          sc = cache()->registerForWrite(trH1, place1, refNamed1.ptr());
          if ( !sc.isSuccess() )   {
            checkError(sc,"Error cache.registerForWrite(trH1, place1, TNamed) [1]");
          }
          sc = cache()->registerForWrite(trH1, place3, refAnother.ptr());
          if ( !sc.isSuccess() )   {
            checkError(sc,"Error cache.registerForWrite(trH1, place3, AnotherTNamed)   [3]");
          }
          sc = cache()->registerForWrite(trH1, place4, refNS.ptr());
          if ( !sc.isSuccess() )   {
            checkError(sc,"Error cache.registerForWrite(trH1, place4, A::B::C::Named) [4]");
          }
        }
      }

      std::vector<const Token*> conts;
      sc = exp->containers(dsc1, conts);
      for(std::vector<const Token*>::iterator k2=conts.begin(); k2 != conts.end(); ++k2)  {
        if ( (*k2)->contID() == "test_cont_1" )  {
          std::cout << " ============== SELECT on Directory:" 
                    << (*k2)->contID() << " ==========" << std::endl;
          Token* tok = 0;
          DbSelect sel("");
          std::string cntName = (*k2)->contID();
          cache()->setPrintout(1);
          sc = exp->select(dsc1, cntName, sel);
          checkError(sc,"select(dsc1, cntName, sel)");

          size_t num_token = 0, num_object = 0;
          while ( sc.isSuccess() )  {
            sc = exp->next(sel,tok);
            checkError(sc,"next(sel,tok)");
            if ( sc.isSuccess() )  {
              Ref<TNamed> ref(cache(), tok);
              if ( ref )  {
                const char* name = ref->GetTitle();
                if ( name )  {
                  num_object++;
                }
              }
              ++num_token;
              if ( num_token < 20 || (num_token > 1200 && num_token < 1300) )  {
                cache()->setPrintout(1);
                std::cout << tok->toString() << std::endl
                          << "    Type: " << typeid(*(ref.ptr())).name()
                          << std::endl;
              }
              else {
                cache()->setPrintout(0);
              }
            }
            else  {
              std::cout << "End of token loop!" << std::endl;
            }
          }
          std::cout << "Successfully received " << num_token  << " tokens." << std::endl;
          std::cout << "Successfully received " << num_object << " objects." << std::endl;
          cache()->clear();
        }
      }
      std::cout << "Commit transaction..." << std::endl;
      sc = cache()->rollback(trH1);
      checkError(sc,"Error cache.commit(...) [1]");

      std::cout << " ========================================" << std::endl;
      cache()->disconnect(0);
      return sc;
    }

  };
}

pool::TestRunner<pool::ObjectIterator> s_testObjectIterator("ObjectIterator");
