#ifndef COLLECTIONBASE_TOKENLIST_H
#define COLLECTIONBASE_TOKENLIST_H

#include <string>
#include <map>
#include <iosfwd>


namespace pool {

  class TokenListSpecification;
  class Token;

  /**
   * @class TokenList TokenList.h CollectionBase/TokenList.h
   *
   * A container class to facilitate the use of Token objects.
   */
  class TokenList
  {
  public:
    /// Default constructor.
    TokenList();

    /**
     * Constructor that takes a Token list specification as input.
     *
     * @param tokenSpec Token list specification.
     * @param sharedTokenSpec Flag set to true if Token list specification is shared externally. 
     */
    TokenList( const TokenListSpecification& tokenSpec, bool sharedTokenSpec = false );

    /** 
     * Copy constructor. Copies both the data and the Token list specification.
     *
     * @param rhs Token list to copy.
     */
    TokenList( const TokenList& rhs );

    /// Default destructor.
    ~TokenList();

    /**
     * Assignment operator. Copies both the data and the Token list specification.
     *
     * @param rhs Token list to copy.
     */
    TokenList& operator=( const TokenList& rhs );

    /**
     * Equality operator. Only compares the Token values, not their names.
     *
     * @param rhs The Token list to compare to.
     */
    bool operator==( const TokenList& rhs ) const;

    /**
     * Inequality operator. Only compares the Token values, not their names.
     *
     * @param rhs The Token list to compare to.
     */
    bool operator!=( const TokenList& rhs ) const;

    /**
     * Extends the Token list specification by one element. If the existing Token list 
     * specification is shared then a copy of this specification and the existing data is 
     * made and the reference to the original specification is dropped before the
     * specification is extended.
     *
     * @param name Token name added to Token list specification. 
     */
    void extend( const std::string& name );

    /**
     * Copies the data of another Token list. The Token list to be copied may contain more
     * elements than the destination Token list. In that case, only the first n Tokens
     * will be copied, where n corresponds to the number of elements in the destination
     * Token list specification.
     *
     * @param rhs Token list from which to copy data.
     */ 
    void copyData( const TokenList& rhs );

    /**
     * Merges into and shares the data of another Token list.
     *
     * @param rhs The Token list whose data is to be shared.
     */
    TokenList& merge( TokenList& rhs );

    /// Returns the size of the Token list.
    size_t size() const;

    /**
     * Returns a reference to a metadata Token given its name.
     *
     * @param name Name of Token.
     */
    Token& operator[]( const std::string& name );

    /**
     * Returns a constant reference to a metadata Token given its name.
     *
     * @param name Name of Token.
     */
    const Token& operator[]( const std::string& name ) const;

    /** 
     * Returns a reference to a metadata Token given its index.
     * 
     * @param name Index of Token.
     */
    Token& operator[]( unsigned int index );

    /** 
     * Returns a constant reference to a metadata Token given its index.
     * 
     * @param name Index of Token.
     */
    const Token& operator[]( unsigned int index ) const;

    /**
     * Outputs the Token names and associated values to a specified stream.
     *
     * @param os Stream to which Token information is to be output.
     */
    std::ostream& toOutputStream( std::ostream& os ) const;

    /// Base class for forward iterators used to navigate TokenList objects.
    class iterator_base 
    {
    public:
      bool operator==( const iterator_base& rhs ) const { return m_iterator == rhs.m_iterator; }
      bool operator!=( const iterator_base& rhs ) const { return m_iterator != rhs.m_iterator; }
      void operator++() { ++m_iterator; }
    protected:
      iterator_base( std::map< std::string, Token* >::const_iterator theIterator ) 
        : m_iterator( theIterator ) {}
      iterator_base( const iterator_base& rhs )
        : m_iterator( rhs.m_iterator ) {}
      iterator_base& operator=( const iterator_base& rhs ) 
      { 
        m_iterator = rhs.m_iterator; 
        return *this; 
      }
      virtual ~iterator_base() {}
      std::map< std::string, Token* >::const_iterator m_iterator;      
    };

    /// Forward iterator class for navigation of TokenList objects.
    class iterator : public iterator_base
    {
    public:
      ~iterator() {};
      iterator( const iterator_base& rhs )
        : iterator_base( rhs ) {}
      iterator& operator=( const iterator_base& rhs ) 
      { 
        iterator_base::operator=( rhs );
        return *this; 
      }
    private:
      friend class TokenList;
      iterator( std::map< std::string, Token* >::const_iterator theIterator )
        : iterator_base( theIterator ) {}
    public:
      Token* operator->() { return const_cast<Token*>( m_iterator->second ); }
      Token& operator*() { return const_cast<Token&>( *( m_iterator->second ) ); }
      iterator& operator++() { ++m_iterator; return *this; }
 
      const std::string& tokenName() const { return m_iterator->first; }
    };

    /// Returns a forward iterator pointing to first element in Token list.
    iterator begin();

    /// Returns a forward iterator pointing to last element in Token list.
    iterator end();

    /// Constant forward iterator class for navigation of TokenList objects.
    class const_iterator : public iterator_base
    {
    public:
      ~const_iterator() {};
      const_iterator( const iterator_base& rhs ): iterator_base( rhs ) {}
      const_iterator& operator=( const iterator_base& rhs ) 
      { 
        iterator_base::operator=( rhs ); 
        return *this; 
      }
    private:
      friend class TokenList;
      const_iterator( std::map< std::string, Token* >::const_iterator theIterator )
        : iterator_base( theIterator ) {}
    public:
      const Token* operator->() const { return m_iterator->second; }
      const Token& operator*() const { return *( m_iterator->second ); }
      const_iterator& operator++() { ++m_iterator; return *this; }
      const std::string& tokenName() const { return m_iterator->first; }
    };

    /// Returns a constant forward iterator pointing to first element in Token list.
    const_iterator begin() const;

    /// Returns a constant forward iterator pointing to last element in Token list.
    const_iterator end() const;

  private:
    /// Token list specification.
    TokenListSpecification* m_tokenListSpec;

    /// Flag indicating whether Token list specification is shared (DEFAULT is false).
    bool m_ownTokenListSpec;

    /// Map of Tokens using Token name as key.
    std::map< std::string, Token* > m_tokenForName;

    /// Map of Tokens using position in map as key.
    std::map< int, Token* > m_tokenForIndex;
  };
}


// Inline methods.
inline size_t
pool::TokenList::size() const
{
  return m_tokenForName.size();
}

inline bool
pool::TokenList::operator!=( const pool::TokenList& rhs ) const
{
  return ( ! ( *this == rhs ) );
}

inline pool::TokenList::iterator
pool::TokenList::begin()
{
  return pool::TokenList::iterator( m_tokenForName.begin() );
}

inline pool::TokenList::iterator
pool::TokenList::end()
{
  return pool::TokenList::iterator( m_tokenForName.end() );
}

inline pool::TokenList::const_iterator
pool::TokenList::begin() const
{
  return pool::TokenList::const_iterator( m_tokenForName.begin() );
}

inline pool::TokenList::const_iterator
pool::TokenList::end() const
{
  return pool::TokenList::const_iterator( m_tokenForName.end() );
}

#endif

