#ifndef COLLECTIONBASE_TOKENLISTSPECIFICATION_H
#define COLLECTIONBASE_TOKENLISTSPECIFICATION_H

#include <string>
#include <typeinfo>
#include <vector>
#include <map>

namespace pool {

  /**
   * @class TokenListSpecification TokenListSpecification.h CollectionBase/TokenListSpecification.h
   *
   * A specification of the metadata Tokens to be used by a collection.
   */
  class TokenListSpecification 
  {
  public:
    /// Default Constructor.
    TokenListSpecification();

    /**
     * Copy constructor.
     *
     * @param rhs Token list specification to be copied.
     */
    TokenListSpecification( const TokenListSpecification& rhs );

    /**
     * Equality operator.
     *
     * @param rhs Token list specification to be compared to.
     */
    bool operator==( const TokenListSpecification& rhs ) const;

    /** 
     * Inequality operator.
     *
     * @param rhs Token list specification to be compared to.
     */
    bool operator!=( const TokenListSpecification& rhs ) const;

    /// Decrements the counter of references to this object.
    void release() const;

    /**
     * Adds a new Token name to the Token list specification.
     *
     * @param name Name of new Token.
     */
    void extend( const std::string& name );

    /// Returns the size of the Token list specification.
    size_t size() const;

    /**
     * Returns the index of a Token in the Token list specification, given its name. 
     * If the name is not found a value of -1 is returned.
     *
     * @param name Name of Token.
     */
    int tokenIndex( const std::string& tokenName ) const;

    /**
     * Returns the name of a Token given its index in the Token list specification.
     *
     * @param index Index of Token.
     */
    const std::string& tokenName( int tokenIndex ) const;

    /**
     * Returns the name of a Token given its index in the Token list specification.
     *
     * @param index Index of Token.
     */
    const std::string& operator[]( size_t tokenIndex ) const;

    /// Iterator class for navigation over TokenListSpecification objects.
    class const_iterator
    {
    public:
      ~const_iterator() {};

      const_iterator( const const_iterator& rhs );
      const_iterator& operator=( const const_iterator& rhs );

    private:
      friend class TokenListSpecification;
      const_iterator( std::vector< std::string >::const_iterator theIterator );

    public:
      const std::string* operator->() const;
      const std::string& operator*() const;
      void operator++();
      bool operator==( const const_iterator& rhs ) const;
      bool operator!=( const const_iterator& rhs ) const;

    private:
      std::vector< std::string >::const_iterator m_iterator;
    };

    /// Returns a forward iterator
    const_iterator begin() const;
    const_iterator end() const;

  protected:
    /// Friend declaration of TokenList class to allow it to call addRef method.
    friend class TokenList;

    /// Increments counter of references to this object.
    void addRef() const;

    /// Destructor.
    ~TokenListSpecification();

    /// No assignment operator.
    TokenListSpecification& operator=( const TokenListSpecification& );

  private:
    /// Counter of references to this object.
    mutable int m_refCounter;

    /// Token names used by Token list specification.
    std::vector< std::string > m_tokenNames;

    /// Map of Token names to Token indices.
    std::map< std::string, int >  m_indexForName;
  };
}


inline bool
pool::TokenListSpecification::operator!=( const pool::TokenListSpecification& rhs ) const
{
  return ( ! ( *this == rhs ) );
}

inline size_t
pool::TokenListSpecification::size() const
{
  return m_tokenNames.size();
}

inline const std::string&
pool::TokenListSpecification::operator[]( size_t tokenIndex ) const
{
  return this->tokenName( static_cast<int>( tokenIndex ) );
}

inline
pool::TokenListSpecification::const_iterator::const_iterator( const pool::TokenListSpecification::const_iterator& rhs ):
  m_iterator( rhs.m_iterator )
{}
  
inline pool::TokenListSpecification::const_iterator&
pool::TokenListSpecification::const_iterator::operator=( const pool::TokenListSpecification::const_iterator& rhs )
{
  m_iterator = rhs.m_iterator;
  return *this;
}

inline
pool::TokenListSpecification::const_iterator::const_iterator( std::vector< std::string>::const_iterator theIterator )
  : m_iterator( theIterator )
{}

inline const std::string*
pool::TokenListSpecification::const_iterator::operator->() const
{
  return &*m_iterator;
}
  
inline const std::string&
pool::TokenListSpecification::const_iterator::operator*() const
{
  return *m_iterator;
}

inline void
pool::TokenListSpecification::const_iterator::operator++()
{
  m_iterator++;
}

inline bool
pool::TokenListSpecification::const_iterator::operator==( const pool::TokenListSpecification::const_iterator& rhs ) const
{
  return m_iterator == rhs.m_iterator;
}

inline bool
pool::TokenListSpecification::const_iterator::operator!=( const pool::TokenListSpecification::const_iterator& rhs ) const
{
  return m_iterator != rhs.m_iterator;
}

inline pool::TokenListSpecification::const_iterator
pool::TokenListSpecification::begin() const
{
  return pool::TokenListSpecification::const_iterator( m_tokenNames.begin() );
}

inline pool::TokenListSpecification::const_iterator
pool::TokenListSpecification::end() const
{
  return pool::TokenListSpecification::const_iterator( m_tokenNames.end() );
}

#endif
