#ifndef POOL_CACHESVC_H
#define POOL_CACHESVC_H

// Include files
#include "DataSvc/ICacheSvc.h"
#include "DataSvc/ObjectDeletePolicy.h"
#include <set>
#include "RVersion.h"

namespace pool    {
  class ReferenceHandler;
  class ICache;
  class IPersistencySvc;

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
typedef ROOT::Reflex::Type TypeH;
#else
typedef Reflex::Type TypeH;
#endif

  /** @class DataSvc DataSvc.h src/DataSvc.h
      DataSvc is the abstract interface for retrieving/registering 
      persitable objects using the storage services.

      @author Giacomo Govi
      @date 15/07/2002
  */
  class CacheSvc : public pool::ICacheSvc {

    public:

      /** constructor with an already created PersistencySvc
      */
      CacheSvc(IPersistencySvc*, const ObjectDeletePolicy&);

      /// Standard destructor
      virtual ~CacheSvc();

    // retrieve the object pointer from a given cache line
    const SharedAnyPtr& getObject(CacheEntry* aCacheEntry);
 
    // reset (delete) the object pointer of a specified cache line 
    bool resetObject(CacheEntry* aCacheEntry);
  
    // check if the object in a specified cache line has been loaded
    bool isObjectLoaded(const CacheEntry* aCacheEntry);

    // retrieve the token from a given cache line
    Token* getToken(CacheEntry* aCacheEntry);

    // register a cacheline to be submitted to write procedure
    bool registerForWrite(const Placement& place,
				      CacheEntry* aCacheEntry);

    // register a cacheline to be submitted to update procedure
    bool registerForUpdate(CacheEntry* aCacheEntry);

    // register a cacheline to be submitted to delete procedure
    bool registerForDelete(CacheEntry* aCacheEntry);

    // set up a reference to the specified cache line, with type checking
    bool linkCacheEntry(CacheEntry* aCacheEntry);

    // set up a reference to the specified cache line, with type checking
    bool linkCacheEntry(CacheEntry* aCacheEntry,
                       const TypeH& clientClass);

    // remove the reference to the specified cache line
    bool unlinkCacheEntry(CacheEntry* aCacheEntry);

    // retrieve the cache line for a specified token. 
    CacheEntry* registerToken(const Token& aToken);

    // retrieve the cache line for a specified object pointer. 
    CacheEntry* registerObject( const SharedAnyPtr& anObjPtr );

    // add a specified cache line in the cache
    bool registerCacheEntry( CacheEntry *aCacheEntry); 

    // retrieve the current delete policy for the objects
    const ObjectDeletePolicy& objectDeletePolicy();

    // clear the cache content
    void clearCache();

    // delete the objects from all the cache entries
    void resetCache();

    // retrive the number of cache entry elements
    unsigned int cacheSize();

    // returns the number of cache entries marked (write,update or delete)
    unsigned int numberOfMarkedObjects();

    // register a client 
    void registerRef( IRegisteredRef* ref );

    // unregister a client 
    void unregisterRef( IRegisteredRef* ref );

    // procedure to be executed at commit time
    bool resetMarkedObjects();

    // procedure to be executed at rollback time
    bool unmarkObjects();

    // dump cache content (debugging)
    void dumpCacheContent();

    // cache mgr instance (debugging)
    ICache* cache();

    protected:

    // execute the loading of the object from the DB to the cache line
    bool _retrieveObject(  CacheEntry* aCacheEntry );
     
    // remove the specified line from the cache
    bool _removeCacheEntry(CacheEntry* aCacheEntry);

  private:

      ObjectDeletePolicy m_deletePolicy;      
      ICache* m_cache;
      IPersistencySvc* m_persSvc;
      ReferenceHandler* m_refHandler;
      std::set< IRegisteredRef* > m_refs;
      CacheSvc();


  };   // class CacheSvc
}      // namespace pool

#endif  // POOL_CACHESVC_H











