#ifndef OBJECTRELATIONALACCESS_OBJECTRELATIONDEFAULTMAPPINGBUILDER_H
#define OBJECTRELATIONALACCESS_OBJECTRELATIONDEFAULTMAPPINGBUILDER_H

#include "ObjectRelationalAccess/ObjectRelationalTableRegister.h"
#include <string>
#include <map>
#include <set>

#include "RVersion.h"

namespace coral {
  class MessageStream;
}

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
namespace ROOT{
  namespace Reflex{
    class Type;
  }
}
using namespace ROOT;
#else
namespace Reflex{
  class Type;
}
#endif 

namespace pool {
  class ObjectRelationalMapping;
  class ObjectRelationalMappingElement;
  /**
     @class ObjectRelationalDefaultMappingBuilder ObjectRelationalDefaultMappingBuilder.h ObjectRelationalAccess/ObjectRelationalDefaultMappingBuilder.h

     Class responsible for the construction of a mapping given the C++ class definitions, following default rules.
  */

  class ObjectRelationalDefaultMappingBuilder {

    public:

    static const size_t MaxColumnsPerTable;

    public:
    /// Constructor
    ObjectRelationalDefaultMappingBuilder();
    /// Constructor
    explicit ObjectRelationalDefaultMappingBuilder(bool arraysInBlobs);
    /// Destructor
    ~ObjectRelationalDefaultMappingBuilder();

    /**
       Processes a class given its name. Returns the success (or not) of the operation.
       @param mapping The mapping to be modified
       @param classDictionary The reflection information of the class to be processed
    */
    bool processClass( ObjectRelationalMapping& mapping,
                       const Reflex::Type& classDictionary );

    /**
       Processes a class as a dependent object. Returns the success (or not) of the operation.
       @param parentClassMapping the structure containing the parent mapping. The new mapping will be appended.
       @param classDictionary The reflection information of the class to be processed
     */
    bool processDependentClass( ObjectRelationalMapping& parentClassMapping,
                                const Reflex::Type& classDictionary );
    
    /**
       Function for processing custom mapping on C-Arrays
     */   
    bool buildCArrayElementTree( ObjectRelationalMappingElement& topElement );
    
    /**
     * Returns the current list of involved tables
     */
    ObjectRelationalTableRegister& tableRegister();

    // Calculate the number of columns required for a given class
    static size_t sizeInColumns(const Reflex::Type& topLevelClassType, bool inlinedCArraysOnly=false);
    
    private:
    
    /// The register of the tables involved in the mapping
    ObjectRelationalTableRegister m_tableRegister;

    coral::MessageStream& m_log;
    /// A flag indicating whether the arrays should be blobified
    bool m_arraysInBlobs;

    private:

    // process Objects
    bool processObject( ObjectRelationalMappingElement& mappingElement,
                        const Reflex::Type& classDictionary );

    // select the required procedure to process the item
    bool processItem( ObjectRelationalMappingElement& parentelement,
                      const std::string& attributeName,
                      const std::string& attributeNameForSchema,
                      const Reflex::Type& attributeType );

    // process Primitives
    bool processPrimitive( ObjectRelationalMappingElement& parentelement,
                           const std::string& attributeName,
                           const std::string& attributeNameForSchema,
                           const Reflex::Type& attributeType );

    // process Blob types
    bool processBlob( ObjectRelationalMappingElement& parentelement,
                      const std::string& attributeName,
                      const std::string& attributeNameForSchema,
                      const Reflex::Type& attributeType );

    // process the most inner element of the tree
    bool processLeafElement( const std::string& elementType,
                             ObjectRelationalMappingElement& parentelement,
                             const std::string& attributeName,
                             const std::string& attributeNameForSchema,
                             const std::string& typeName );

    // process embedded classes
    bool processEmbeddedClass( ObjectRelationalMappingElement& parentelement,
                               const std::string& attributeName,
                               const std::string& attributeNameForSchema,
                               const Reflex::Type& attributeType,
                               bool isRef );

    // process stl containers
    bool processArray( ObjectRelationalMappingElement& parentelement,
                       const std::string& attributeName,
                       const std::string& attributeNameForSchema,
                       const Reflex::Type& attributeType );

    // select the required procedure for C-arrays
    bool processCArray( ObjectRelationalMappingElement& parentelement,
                        const std::string& attributeName,
                        const std::string& attributeNameForSchema,
                        const Reflex::Type& attributeType );

    // process C-arrays as containers
    bool processCArrayItem( ObjectRelationalMappingElement& parentelement,
                            const std::string& attributeName,
                            const std::string& attributeNameForSchema,
                            const Reflex::Type& attributeType,
                            const Reflex::Type& arrayElementType );

    // process C-arrays with inline columns
    bool processInlineCArrayItem( ObjectRelationalMappingElement& parentelement,
                                  const std::string& attributeName,
                                  const std::string& attributeNameForSchema,
                                  const Reflex::Type& attributeType,
                                  const Reflex::Type& arrayElementType );

    // process pool::Ptr objects
    bool processPoolPtr( ObjectRelationalMappingElement& parentelement,
                         const std::string& attributeName,
                         const std::string& attributeNameForSchema,
                         const Reflex::Type& attributeType );

    // process pool::PolyPtr objects
    bool processPoolPolyPtr( ObjectRelationalMappingElement& parentelement,
                             const std::string& attributeName,
                             const std::string& attributeNameForSchema,
                             const Reflex::Type& attributeType );

    // recursive function to calculate the required colums 
    static void _sizeInColumns(const Reflex::Type& typ, size_t& sz, bool inlinedCArraysOnly=false);
  };

}

#endif
