#ifndef OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGPERSISTENCY_H
#define OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGPERSISTENCY_H

#include <map>
#include <vector>
#include <string>
#include <set>

namespace coral {

  class AttributeList;
  class ISchema;
  class ITableDataEditor;
}

namespace pool 
{
  
  class ObjectRelationalMapping;
  class ObjectRelationalMappingElement;

  /**
   * @class ObjectRelationalMappingPersistency ObjectRelationalMappingPersistency.h ObjectRelationalAccess/ObjectRelationalMappingPersistency.h
   *
   * Class respondible for the I/O of object/relational mappings with a relational database.
   */
  
  class ObjectRelationalMappingPersistency
  {
  public:
    /// Constructor
    explicit ObjectRelationalMappingPersistency( coral::ISchema& schema );

    /// Destructor
    ~ObjectRelationalMappingPersistency();

    /**
     * Initialize the service for updating the tables
     */
    void initializeForWriting();

    /**
     * Retrieves the versions of the mappings stored in the database
     */
    std::map<std::string,std::string> retrieveMappingVersions( bool tmpTable=false) const;

    /**
     * Store the provided version list in the database
     */
    bool storeMappingVersions( const std::map<std::string,std::string>& data, bool tmpTable=false) const;

    /**
     * Provides a new version label for the specified class
     */
    std::string newMappingVersionForClass( const std::string& className );
    
    /**
     * Retrieves the versions of the mappings stored in the database
     */
    std::map<std::pair<std::string,std::string>,std::string> retrieveMappingForClassVersions(bool tmpTable=false) const;

    /**
     * Select the mapping version (if any) corresponding to the specified class version and container name.
     */
    std::string selectMappingVersion( const std::string& classVersion, const std::string& containerName) const;

    /**
     * Select the container name using the specified mapping version
     */
    std::string containerForMappingVersion( const std::string& mappingVersion) const;

    /**
     * Select the mapping versions in use by the specified container
     */
    std::vector<std::string> mappingsForContainer( const std::string& containerName ) const;

    /**
     * Store the provided class version map in the database
     */
    bool storeMappingForClassVersions( std::map<std::pair<std::string,std::string>,std::string> data, bool tmpTable=false) const;

    /**
     * Set the specified mapping version for the provided class version and container.
     */
    bool assignMapping( const std::string& classVersion, const std::string& mappingVersion, const std::string& containerName);
    
    /**
     * Delivers a map of the table usage by containers
     */
    std::map<std::string, std::string> containerTableMap();

    /**
     *  Fills the provided map with the mapping data for the specified version
     */
    void retrieveMapping(const std::string& version, std::map<std::pair<std::string,std::string>,std::pair<std::string,std::string> >& mappingStruct) const;
    
    /**
     * Retrieves a mapping of a given version
     */
    ObjectRelationalMapping* retrieveMapping( const std::string& version, bool tmpTable=false  ) const;

    /**
     * Retrieves a mapping from an old version schema
     */
    ObjectRelationalMapping* retrieveMappingFromOldSchema( const std::string& version ) const;

    /**
     * Stores a mapping structure in the database.
     */
    bool storeMapping( const ObjectRelationalMapping& mapping, bool tmpTable=false);
    
    /**
     * Stores a mapping element in the database.
     */
    bool storeMappingElement( const std::string& version,
                              const std::string& elementId,
                              const ObjectRelationalMappingElement& mappingelement );

    /**
     * Stores a mapping structure in the database.
     */
    bool storeMapping( const ObjectRelationalMapping& mapping, const std::string& classVersion, std::string containerName="" );
    
    /**
     * Removes a mapping from the database given its version
     */
    bool removeMapping( const std::string& version );

    /**
     * Creates all the tables needed for the storage of objects corresponding to a mapping.
     */
    bool materializeMapping( const pool::ObjectRelationalMapping& mapping,
                             bool evolve, bool dryRun=false );
    
    /**
     * Creates all the tables needed for the storage of objects corresponding to a mapping element.
     */
    bool materializeMappingElement( const pool::ObjectRelationalMappingElement& mapping,
                                    bool evolve, bool dryRun=false );

    /**
     * Removes the tables materialized from the database given their mapping version
     */
    bool removeMaterializedTables(const std::string& mappingVersion, bool dryRun=false );

    private:
    
    struct MappingDescription {
        MappingDescription():
          variableType(""),
          elementType(""),
          tableName(""),
          columns(){}
        MappingDescription(const MappingDescription& rhs):
          variableType(rhs.variableType),
          elementType(rhs.elementType),
          tableName(rhs.tableName),
          columns(rhs.columns){}
        MappingDescription& operator==(const MappingDescription& rhs){
          variableType = rhs.variableType;
          elementType = rhs.elementType;
          tableName = rhs.tableName;
          columns = rhs.columns; 
          return *this;
        }
        std::string variableType;
        std::string elementType;
        std::string tableName;
        std::vector<std::string> columns;
    }; 
    
    private:
    /**
     * Reference to the relational schema in use
     */
    coral::ISchema& m_schema;

    /**
     * The set of the mappings in the database
     */
    mutable std::set< std::string > m_storedMappings;

    mutable bool m_versionsLoaded;
    
    /**
     *  Temporary cache for the loading of the mapping from db into the proper structures
     */
    mutable std::map< std::string, std::map<std::string, MappingDescription> > m_tempMappingTable;

    private:
    /// Creates if necessary the mapping tables
    void createIfNecessaryTheMappingTables();
    /// Loads the mapping versions
    void loadMappingVersions();
    
    void makeTableList( const ObjectRelationalMappingElement& element,
                        std::set<std::string>& tableList );

    void buildMappingElement( ObjectRelationalMappingElement& mappingElement, const std::string& scopeName ) const;
    
    private:

    // Returns the mapping version for a class whose mapping has been generated automatically
    static std::string mappingVersionForClass( const std::string& className, int iteration );

    /// Stores a mapping element
    static void storeMappingElement( const std::string& elementId,
                                     const pool::ObjectRelationalMappingElement& element,
                                     coral::ITableDataEditor& dataEditor,
                                     coral::AttributeList& dataBuffer );
  };
}

#endif // OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGPERSISTENCY_H
