#include "DatabaseHandler.h"
#include "Container.h"
#include "PersistencySvc/PersistencySvcException.h"
#include "PersistencySvc/IDataTransform.h"
#include "StorageSvc/IStorageSvc.h"
#include "StorageSvc/IStorageExplorer.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/Transaction.h"
#include "StorageSvc/Shape.h"
#include "StorageSvc/DbOption.h"
#include "POOLCore/Token.h"
#include "StorageSvc/DbTypeInfo.h"

#include <memory>

static const Reflex::Type emptyType = Reflex::Type();


pool::PersistencySvc::DatabaseHandler::DatabaseHandler( pool::IStorageSvc& storageSvc,
                                                        pool::IStorageExplorer& storageExplorer,
                                                        pool::Session* session,
                                                        long technology,
                                                        const std::string& fid,
                                                        const std::string& pfn,
                                                        long accessmode ):
  m_storageSvc( storageSvc ),
  m_storageExplorer( storageExplorer ),
  m_session( session ),
  m_transaction( 0 ),
  m_fileDescriptor( fid, pfn ),
  m_technology( technology ),
  m_accessMode( accessmode )
{
  if ( ! m_storageSvc.connect( m_session,
                               m_accessMode,
                               m_fileDescriptor ).isSuccess() ) {
    throw pool::PersistencySvcException( "Could not connect to the database",
                                         "PersistencySvc::DatabaseHandler::DatabaseHandler" );
  }
}


pool::PersistencySvc::DatabaseHandler::~DatabaseHandler()
{
  if ( this->isTransactionActive() ) {
    this->rollBackTransaction();
  }
  m_storageSvc.disconnect( m_fileDescriptor );
}


bool
pool::PersistencySvc::DatabaseHandler::startTransaction()
{
  if ( m_transaction ) return false;
  if ( m_storageSvc.startTransaction( m_fileDescriptor.dbc(),
                                      m_transaction ).isSuccess() ) {
    return true;
  }
  else return false;
}


bool
pool::PersistencySvc::DatabaseHandler::commitTransaction()
{
  bool result = false;
  if ( m_transaction ) {
    if ( m_storageSvc.endTransaction( m_transaction,
                                      pool::Transaction::TRANSACT_COMMIT ).isSuccess() &&
         m_storageSvc.startTransaction( m_fileDescriptor.dbc(),
                                        m_transaction ).isSuccess() &&
         m_storageSvc.endTransaction( m_transaction,
                                      pool::Transaction::TRANSACT_FLUSH ).isSuccess() ) {
      result = true;
    }
    m_transaction = 0;
  }
  return result;
}


bool
pool::PersistencySvc::DatabaseHandler::commitAndHoldTransaction()
{
  bool result = false;
  if ( m_transaction ) {
    if ( m_storageSvc.endTransaction( m_transaction,
                                      pool::Transaction::TRANSACT_COMMIT ).isSuccess() ) {
      m_transaction = 0;
      if ( m_storageSvc.startTransaction( m_fileDescriptor.dbc(),
                                          m_transaction ).isSuccess() ) {
        result = true;
      }
    }
  }
  return result;
}


void
pool::PersistencySvc::DatabaseHandler::rollBackTransaction()
{
  if ( m_transaction ) {
    m_storageSvc.endTransaction( m_transaction, pool::Transaction::TRANSACT_ROLLBACK );
    m_transaction = 0;
  }
}


bool
pool::PersistencySvc::DatabaseHandler::isTransactionActive() const
{
  if ( m_transaction ) return true;
  else return false;
}


std::vector< std::string >
pool::PersistencySvc::DatabaseHandler::containers()
{
  std::vector< std::string > result;
  if ( m_transaction ) {
    std::vector<const pool::Token*> containerTokens;
    m_storageExplorer.containers( m_fileDescriptor, containerTokens );
    for ( std::vector<const pool::Token*>::const_iterator iToken = containerTokens.begin();
          iToken != containerTokens.end(); ++iToken ) {
      result.push_back( (*iToken)->contID() );
    }
  }
  return result;
}


pool::IContainer*
pool::PersistencySvc::DatabaseHandler::container( const std::string& containerName )
{
  if ( m_transaction ) {
    // Check the existence of a given container.
    std::vector< std::string > allContainers = this->containers();
    for ( std::vector< std::string >::const_iterator iName = allContainers.begin();
          iName != allContainers.end(); ++iName ) {
      if ( *iName == containerName ) {
        return new pool::PersistencySvc::Container( m_fileDescriptor,
                                                    m_storageExplorer,
                                                    m_technology,
                                                    containerName );
      }
    }
  }
  return 0;
}


const std::string&
pool::PersistencySvc::DatabaseHandler::pfn() const
{
  return m_fileDescriptor.PFN();
}


const std::string&
pool::PersistencySvc::DatabaseHandler::fid() const
{
  return m_fileDescriptor.FID();
}


long
pool::PersistencySvc::DatabaseHandler::technology() const
{
  return m_technology;
}


long
pool::PersistencySvc::DatabaseHandler::accessMode() const
{
  return m_accessMode;
}


void
pool::PersistencySvc::DatabaseHandler::reconnect( long accessMode )
{
  bool activeTransaction = this->isTransactionActive();
  if ( activeTransaction ) this->rollBackTransaction();
  if ( m_storageSvc.reconnect( m_fileDescriptor, accessMode ).isSuccess() ) {
    m_accessMode = accessMode;
  }
  if ( activeTransaction ) this->startTransaction();
}


pool::Placement
pool::PersistencySvc::DatabaseHandler::placementHint() const
{
  pool::Placement placement;
  placement.setDatabase( m_fileDescriptor.FID(),
                         DatabaseSpecification::FID );
  placement.setTechnology( m_technology );
  return placement;
}


pool::Token*
pool::PersistencySvc::DatabaseHandler::writeObject( pool::IDataTransform& dataTransform,
                                                    const std::string& containerName,
                                                    const Reflex::Type& shapeID,
                                                    long minorTechnology,
                                                    const void* object,
                                                    const Reflex::Type& type,
                                                    pool::DataHandler* handler )
{
  pool::Token* token = 0;
  if ( ! m_transaction ) return token;
  if ( ! object ) return token;

  // Get the persistent type
  Reflex::Type theShape = shapeID;
  if ( theShape == emptyType ) {
    theShape = type;
  }

  // Get the proper callback object
  pool::DataCallBack& callBackObject = dataTransform.transformation( type, theShape );

  // Get the persistent shape.
  const pool::Shape* shape = 0;
  if ( m_storageSvc.getShape( m_fileDescriptor, theShape, shape ) == pool::IStorageSvc::SHAPE_NOT_AVAILIBLE ) {
    m_storageSvc.createShape(m_fileDescriptor, containerName, theShape, shape );
  }

  if ( shape )  {
    std::auto_ptr<pool::DataCallBack> cb( callBackObject.clone() );
    cb->setObject( object );
    cb->setHandler( handler );
    if ( m_storageSvc.allocate( m_transaction,
                                m_fileDescriptor,
                                containerName,
                                minorTechnology,
                                shape, 
                                cb.get(),
                                token ).isSuccess() ) {
      token->setClassID( shape->shapeID() );
      cb.release(); // Pass ownership of the callback object to the storageSvc
    }
  }
  return token;
}


std::pair< void*, Reflex::Type >
pool::PersistencySvc::DatabaseHandler::readObject( pool::IDataTransform& dataTransform,
                                                   const pool::Guid& transientType,
                                                   const pool::Token& token,
                                                   pool::DataHandler* handler )
{
  std::pair< void*, Reflex::Type > result( (void*)0, (Reflex::Type)0 );
  if ( ! m_transaction ) return result;

  // Get the persistent shape
  const pool::Shape* shape = 0;
  if ( m_storageSvc.getShape( m_fileDescriptor, token.classID(), shape ) != pool::IStorageSvc::IS_PERSISTENT_SHAPE )
    return result;

  // Retrieve the shape type
  const pool::DbTypeInfo* dbTypeInfo = dynamic_cast< const pool::DbTypeInfo* >( shape );
  if ( dbTypeInfo == 0 ) return result;
  Reflex::Type shapeType = dbTypeInfo->clazz();

  // Now we have all the information available. Get the callback object.
  pool::DataCallBack* cb = dataTransform.transformation( transientType,
                                                         shapeType ).clone();
  cb->setHandler( handler );
  if ( m_storageSvc.read( m_fileDescriptor, token, cb, shape, result.second ).isSuccess() ) {
    result.first = const_cast<void*>( cb->object() );
  }
  cb->release();

  return result;
}


std::pair< void*, Reflex::Type >
pool::PersistencySvc::DatabaseHandler::readObject( pool::IDataTransform& dataTransform,
                                                   const Reflex::Type& transientType,
                                                   const pool::Token& token,
                                                   pool::DataHandler* handler )
{
  std::pair< void*, Reflex::Type > result( (void*)0, (Reflex::Type)0 );
  if ( ! m_transaction ) return result;

  // Get the persistent shape
  const pool::Shape* shape = 0;
  if ( m_storageSvc.getShape( m_fileDescriptor, token.classID(), shape ) != pool::IStorageSvc::IS_PERSISTENT_SHAPE )
    return result;

  // Retrieve the shape type
  const pool::DbTypeInfo* dbTypeInfo = dynamic_cast< const pool::DbTypeInfo* >( shape );
  if ( dbTypeInfo == 0 ) return result;
  Reflex::Type shapeType = dbTypeInfo->clazz();

  // Now we have all the information available. Get the callback object.
  pool::DataCallBack* cb = dataTransform.transformation( transientType,
                                                         shapeType ).clone();
  cb->setHandler( handler );
  if ( m_storageSvc.read( m_fileDescriptor, token, cb, shape, result.second ).isSuccess() ) {
    result.first = const_cast<void*>( cb->object() );
  }
  cb->release();

  return result;
}


std::pair< void*, Reflex::Type >
pool::PersistencySvc::DatabaseHandler::readObject( pool::IDataTransform& dataTransform,
                                                   const pool::Token& token,
                                                   pool::DataHandler* handler )
{
  std::pair< void*, Reflex::Type > result( (void*)0, (Reflex::Type)0 );
  if ( ! m_transaction ) return result;

  // Get the persistent shape
  const pool::Shape* shape = 0;
  if ( m_storageSvc.getShape( m_fileDescriptor, token.classID(), shape ) != pool::IStorageSvc::IS_PERSISTENT_SHAPE )
    return result;

  // Retrieve the shape type
  const pool::DbTypeInfo* dbTypeInfo = dynamic_cast< const pool::DbTypeInfo* >( shape );
  if ( dbTypeInfo == 0 ) return result;
  Reflex::Type shapeType = dbTypeInfo->clazz();

  // Now we have all the information available. Get the callback object.
  pool::DataCallBack* cb = dataTransform.transformation( shapeType, shapeType ).clone();
  cb->setHandler( handler );
  if ( m_storageSvc.read( m_fileDescriptor, token, cb, shape, result.second ).isSuccess() ) {
    result.first = const_cast<void*>( cb->object() );
  }
  cb->release();

  return result;
}


bool
pool::PersistencySvc::DatabaseHandler::updateObject( pool::IDataTransform& dataTransform,
                                                     const void* object,
                                                     const Reflex::Type& type,
                                                     const pool::Token& token,
                                                     pool::DataHandler* handler )
{
  bool result = false;
  if ( ! m_transaction ) return result;
  if ( ! object ) return result;

  // Get the persistent shape
  const pool::Shape* shape = 0;
  if ( m_storageSvc.getShape( m_fileDescriptor, token.classID(), shape ) != pool::IStorageSvc::IS_PERSISTENT_SHAPE )
    return result;

  // Retrieve the shape type
  const pool::DbTypeInfo* dbTypeInfo = dynamic_cast< const pool::DbTypeInfo* >( shape );
  if ( dbTypeInfo == 0 ) return result;
  Reflex::Type shapeType = dbTypeInfo->clazz();

  pool::DataCallBack& callBackObject = dataTransform.transformation( type, shapeType );

  std::auto_ptr<pool::DataCallBack> cb( callBackObject.clone() );
  cb->setObject( object );
  cb->setHandler( handler );

  if ( m_storageSvc.update( m_transaction, m_fileDescriptor, shape, cb.get(), token ).isSuccess() ) {
    cb.release(); // Pass ownership of the callback object to the storageSvc
    result = true;
  }

  return result;
}


bool
pool::PersistencySvc::DatabaseHandler::destroyObject( const pool::Token& token )
{
  bool result = false;
  if ( ! m_transaction ) return result;
  if ( m_storageSvc.destroy( m_transaction,
                             m_fileDescriptor,
                             token ).isSuccess() ) {
    result = true;
  }
  return result;
}


bool
pool::PersistencySvc::DatabaseHandler::attribute( const std::string& attributeName,
                                                  void* data,
                                                  const std::type_info& typeInfo,
                                                  const std::string& option ) const
{
  pool::DbOption databaseOption( attributeName, option );
  pool::DbStatus sc = m_storageExplorer.getDatabaseOption( m_fileDescriptor, databaseOption );
  if ( !sc.isSuccess() ) return false;
  if ( databaseOption.i_getValue( typeInfo, data ).isSuccess() ) {
    return true;
  }
  else {
    return false;
  }
}


bool
pool::PersistencySvc::DatabaseHandler::setAttribute( const std::string& attributeName,
                                                     const void* data,
                                                     const std::type_info& typeInfo,
                                                     const std::string& option )
{
  pool::DbOption databaseOption( attributeName, option );
  pool::DbStatus sc = databaseOption.i_setValue( typeInfo, const_cast<void*>( data ) );
  if ( !sc.isSuccess() ) return false;
  sc = m_storageExplorer.setDatabaseOption( m_fileDescriptor,
                                            databaseOption );
  if ( !sc.isSuccess() ) {
    return false;
  }
  else {
    return true;
  }
}


std::set< std::string >
pool::PersistencySvc::DatabaseHandler::parameters() const
{
  std::set< std::string > result;
  std::vector< std::pair<std::string, std::string> > parameters;
  if ( m_storageExplorer.dbParams( m_fileDescriptor, parameters ).isSuccess() ) {
    for ( std::vector< std::pair<std::string, std::string> >::const_iterator iParameter = parameters.begin();
          iParameter != parameters.end(); ++iParameter ) {
      result.insert( iParameter->first );
    }
  }
  return result;
}

std::string
pool::PersistencySvc::DatabaseHandler::parameterValue( const std::string& name ) const
{
  std::string value = "";
  m_storageExplorer.dbParam( m_fileDescriptor, name, value );
  return value;
}

void
pool::PersistencySvc::DatabaseHandler::addParameter( const std::string& name,
                                                     const std::string& value )
{
  m_storageExplorer.addDbParam( m_fileDescriptor, name, value );
}
