#include "ObjectReader.h"
#include "ObjectRelationalReadDataBuffer.h"
#include "MappingManager.h"
#include "IContainerHandler.h"
#include "RelationalStorageServiceNames.h"
#include "RelationalStorageServiceException.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeListSpecification.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Blob.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/DbDataHandlerGuard.h"
#include "POOLCore/Ptr.h"
#include "POOLCore/Reference.h"
#include "POOLCore/Token.h"
#include "Reflex/Reflex.h"
#include <memory>
#include <sstream>

namespace pool {
  namespace RelationalStorageService {
    union U_Primitives {
      bool b;
      char c;
      unsigned char uc;
      short s;
      unsigned short us;
      int i;
      unsigned int ui;
      long l;
      unsigned long ul;
      float f;
      double d;
      long double ld;
    };
    
    class RelationalPtrLoader: virtual public pool::IPtrLoader {
      public:
        RelationalPtrLoader(ObjectReader& reader,
                            ObjectRelationalReadDataBuffer& buffer,
                            void* objectData,
                            int oid,
                            bool anyObject,
                            pool::DataCallBack& callback ):
          m_isValid(true),
          m_reader(reader),
          m_buffer(buffer),
          m_objectData(objectData),
          m_oid(oid),
          m_anyObject(anyObject),
          m_callback(callback){}

        virtual ~RelationalPtrLoader(){}
        
        void* load() const {
          void* ret = 0;
          //if(m_isValid) ret = m_reader.loadPoolPointer(m_buffer,m_objectData,m_oid,m_anyObject,m_callback);
          if(m_isValid) ret = m_reader.loadPoolPointer(m_buffer,m_oid,m_anyObject,m_callback);
          return ret;
        }

        void notify() const {
          if(m_isValid) m_reader.eraseLoader(this);
        }

        void invalidate() {
          m_isValid = false;
        }

        bool isValid() const {
          return m_isValid;
        }

      private:
        bool m_isValid;
        ObjectReader& m_reader;
        ObjectRelationalReadDataBuffer& m_buffer;
        void* m_objectData;
        int m_oid;
        bool m_anyObject;
        pool::DataCallBack& m_callback;
  };
  
  }
}


pool::RelationalStorageService::ObjectReader::ObjectReader( coral::ISchema& schema,
                                                            const TypeH& dictionaryEntry,
                                                            MappingManager& mappingManager,
                                                            pool::DbDatabase* databaseHandle ):
  DbIOHandler(databaseHandle),
  m_schema(schema),
  m_dataBuffer( 0 ),
  m_refDict( TypeH::ByTypeInfo( typeid(pool::Reference) ) ),
  m_mappingManager(mappingManager),
  m_poolPointerLoaders()
{
  m_dataBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_context);
  const ObjectRelationalMappingElement& mappingElement = m_mappingManager.mappingForClass( dictionaryEntry ).element();
  m_dataBuffer->build( schema, mappingElement, dictionaryEntry );
}

pool::RelationalStorageService::ObjectReader::~ObjectReader()
{
  delete m_dataBuffer;
  for(std::set<pool::IPtrLoader*>::iterator iPtr=m_poolPointerLoaders.begin();
      iPtr!=m_poolPointerLoaders.end();
      ++iPtr){
    (*iPtr)->invalidate();
  }
}

bool
pool::RelationalStorageService::ObjectReader::readObject( void* objectData,
                                                          int oid,
                                                          bool anyObject,
                                                          pool::DataCallBack& callback )
{
  std::auto_ptr< coral::IQuery > query( makeQuery(*m_dataBuffer,oid,anyObject) );
  coral::ICursor& cursor = query->execute();

  if ( cursor.next() ) {
    
    readObjectData( cursor,*m_dataBuffer,objectData,oid,callback );
    processDependentBuffers( *m_dataBuffer, objectData, oid, anyObject, callback);
    
    return true;
  }
  else {
    return false;
  }
}

void*
pool::RelationalStorageService::ObjectReader::loadPoolPointer( ObjectRelationalReadDataBuffer& buffer,
                                                               int oid,
                                                               bool anyObject,
                                                               pool::DataCallBack& callback ){
  void* ptrData = 0;
  std::auto_ptr< coral::IQuery > query(makeQuery( buffer,oid,anyObject));
  coral::ICursor& cursor = query->execute();
  if ( cursor.next() ) {
    Reflex::Type ptrType;
    if( buffer.mapping().elementType()==pool::ObjectRelationalMappingElement::PoolPointer ){
      ptrType = buffer.dictionary().TemplateArgumentAt(0);
    } else if ( buffer.mapping().elementType()==pool::ObjectRelationalMappingElement::DependentObject ){
      ptrType = buffer.dictionary();
    }
    if(!ptrType){
        throw pool::RelationalStorageService::RelationalStorageServiceException(
          "Could not resolve the pool pointer class.",
          "ObjectReader::loadPoolPointer" );
    }
    
    ptrData = ptrType.Allocate();
    ptrType.Construct(ROOT::Reflex::Type(0,0),std::vector<void*>(),ptrData);
    readObjectData( cursor,buffer,ptrData,oid,callback );
    processDependentBuffers( buffer, ptrData, oid, anyObject, callback);
  }
  return ptrData;
}

bool pool::RelationalStorageService::ObjectReader::eraseLoader( const IPtrLoader* loader ){
  return m_poolPointerLoaders.erase(const_cast<IPtrLoader*>(loader));
}

bool
pool::RelationalStorageService::ObjectReader::readArray( const std::vector<int>& identity,
                                                         ObjectRelationalReadDataBuffer& buffer,
                                                         void* address,
                                                         pool::DataCallBack& callback )
{
  std::auto_ptr< coral::IQuery > query( buffer.table().newQuery() );
  const coral::AttributeListSpecification& outputSpecification = buffer.dataSpecification();
  for (coral::AttributeListSpecification::const_iterator iSp=outputSpecification.begin(); iSp != outputSpecification.end(); ++iSp ) {
    query->addToOutputList( iSp->name() );
    query->defineOutputType( iSp->name(),iSp->typeName());
  }

  coral::AttributeListSpecification::const_iterator iSpec = outputSpecification.begin();
  coral::AttributeList& whereData = buffer.whereDataForId();
  coral::AttributeList::iterator iData = whereData.begin();

  if( buffer.mapping().isDependentTree() ) {
    // skip the refid column for dependent objects
    ++iSpec;
    // set the identity for the refid
    iData->data<std::string>() = buffer.metadata()[0];
    ++iData;
  }
  
  query->addToOrderList( iSpec->name() ); // Order by pos.

  // Fill-in the identities.  
  for ( std::vector<int>::const_iterator id = identity.begin();
        id != identity.end(); ++id ) {
    iData->data<int>() = *id;
    ++iData;
  }

  // Set the condition
  query->setCondition( buffer.whereClauseForId(), whereData );

  // Prefetch rows.
  query->setRowCacheSize( 100 ); // We should better define this value !!!

  // Create the identity for the sub-elements
  std::vector<int> identityForSubElements( 1 + identity.size(), 0 );

  // Check if the container is an associative one
  TypeH keyType;
  TypeH dictionary = buffer.dictionary();
  TypeH iteratorDereferenceReturnType = buffer.containerHandler().iteratorReturnType();
  pool::RelationalStorageService::U_Primitives primitiveStub;
  int keyItems = buffer.keyItems();
  
  Reflex::Member firstField;
  if ( keyItems > 0 ) {
    keyType = dictionary.TemplateArgumentAt(0);
    if(keyType){
      firstField = iteratorDereferenceReturnType.MemberByName( "first" );
      if ( ! firstField ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException(
          "Could not retrieve the \"first\" field for the class \"" +
          iteratorDereferenceReturnType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
          "ObjectReader::readArray" );
      }
    }
  }
  
  // Issue the query...
  coral::ICursor& cursor = query->execute();

  bool isArray = dictionary.IsArray();
  bool isElementFundamental = iteratorDereferenceReturnType.IsFundamental();

  size_t arraySize = dictionary.ArrayLength();
  
  unsigned int i=0;
  while ( cursor.next() ) {

    const coral::AttributeList& row = cursor.currentRow();
    
    // Get the Identity
    coral::AttributeList::const_iterator iAttribute = row.begin();
    if( buffer.mapping().isDependentTree() ) {
      // skip the refid column
      ++iAttribute;
    }
    int arrayIndex = iAttribute->data< int >();
    for ( std::vector<int>::iterator idout = identityForSubElements.begin();
          idout != identityForSubElements.end(); ++idout ) {
      *idout = iAttribute->data< int >();
      ++iAttribute;
    }

    // Create a new element for the array
    void* objectData = 0;
    if( isArray ){
      if(arrayIndex < (int)arraySize){
        // the memory has been allocated already!
        objectData = static_cast<char*>(address)+arrayIndex*iteratorDereferenceReturnType.SizeOf();
        if(!isElementFundamental){
          // in this case the initialization is required: use default constructor...
          iteratorDereferenceReturnType.Construct(TypeH(0,0),std::vector< void* >(),objectData);
        }
      } else {
        // not sure what to do in this case... exception or ignore???
        continue;
      }
    } else {
      if(isElementFundamental){
        objectData = &primitiveStub;
      } else {
        objectData = iteratorDereferenceReturnType.Construct().Address();
      }
    }

    void* componentData = objectData;
    void* keyData = 0;

    // Copy the data
    std::vector<std::pair<int,int> >::const_iterator iOffset = buffer.offsets().begin();

    if ( keyType ) { // treat the key object first
      keyData = static_cast< char* >( objectData ) + firstField.Offset();
      for ( int i = 0; i < keyItems; ++i ) {
        int offset = iOffset->first;
        if ( offset < 0 ) {
          pool::Token::OID_t oid;
          oid.first = iAttribute->data<int>();
          ++iAttribute;
          ++iOffset;
          oid.second = iAttribute->data<int>();
          void* refAddress = static_cast< char* >( keyData ) - ( 1 + offset );
          // Call here the ref handler to get the oid...
          pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
          callback.notify( pool::DataCallBack::GET, pool::DataCallBack::REFERENCE, m_refDict, refAddress, &oid );
        }
        else {
          if( iAttribute->specification().type() == typeid(coral::Blob) ){
            buffer.readBlob( iOffset->first, iAttribute->data<coral::Blob>(), static_cast< char* >( keyData ) + iOffset->first );
          } else {
            iAttribute->copyValueToAddress( static_cast< char* >( keyData ) + iOffset->first );
          }
        }
        ++iAttribute;
        ++iOffset;
      }
      const Reflex::Member secondField = iteratorDereferenceReturnType.MemberByName( "second" );
      if ( ! secondField ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not retrieve the \"second\" field for the class \"" +
                                                                                   iteratorDereferenceReturnType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                 "ObjectReader::readArray" );
      }
      componentData = static_cast< char* >( objectData ) + secondField.Offset();
    }

    while ( iAttribute != row.end() ) {
      int offset = iOffset->first;
      if ( offset < 0 ) {
        pool::Token::OID_t oid;
        oid.first = iAttribute->data<int>();
        ++iAttribute;
        ++iOffset;
        oid.second = iAttribute->data<int>();
        void* refAddress = static_cast< char* >( componentData ) - ( 1 + offset );
        // Call here the ref handler to get the oid...
        pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
        callback.notify( pool::DataCallBack::GET, pool::DataCallBack::REFERENCE, m_refDict, refAddress, &oid );
      }
      else {
        if( iAttribute->specification().type() == typeid(coral::Blob) ){
          buffer.readBlob( iOffset->first, iAttribute->data<coral::Blob>(), static_cast< char* >( componentData ) + iOffset->first );
        } else {
          iAttribute->copyValueToAddress( static_cast< char* >( componentData ) + iOffset->first );
        }
      }
      ++iAttribute;
      ++iOffset;
    }

    // Loop over the arrays now...
    for ( pool::RelationalStorageService::ObjectRelationalReadDataBuffer::iterator iBuffer = buffer.begin();
          iBuffer != buffer.end(); ++iBuffer ) {
      void* theAddress = componentData;
      if ( (*iBuffer)->isKeyData() ) {
        theAddress = keyData;
      }
      if ( ! this->readArray( identityForSubElements, **iBuffer,
                                static_cast< char* >( theAddress ) + (*iBuffer)->offsetToParent(),
                              callback ) ) return false;
    }

    if(!isArray) {
      size_t prevSize = buffer.containerHandler().size( address );
      buffer.containerHandler().appendNewElement( address, objectData );
      bool inserted = buffer.containerHandler().size( address )>prevSize;
      if ( ! ( iteratorDereferenceReturnType.IsFundamental() ) ) {
        iteratorDereferenceReturnType.Destruct( objectData );
      }
      if ( !inserted ) return false;
    }
    ++i;
  }

  buffer.containerHandler().finalize( address );
  
  return true;
}

bool
pool::RelationalStorageService::ObjectReader::readPoolPointer( ObjectRelationalReadDataBuffer& dataBuffer,
                                                               void* objectData,
                                                               int oid,
                                                               bool anyObject,
                                                               pool::DataCallBack& callback ){
  ROOT::Reflex::Type poolPtrType = dataBuffer.dictionary();
  ROOT::Reflex::Object poolPtrObj(poolPtrType,objectData);
  poolPtrObj.Invoke("reset");
  
  ROOT::Reflex::Member ptrMember = poolPtrType.MemberByName("m_loader");
  RelationalPtrLoader* loader = new RelationalPtrLoader(*this,dataBuffer,objectData,oid,anyObject,callback);
  *(void**)(static_cast<char*>(objectData)+ptrMember.Offset())=loader;
  m_poolPointerLoaders.insert(loader);
  return true;
}

bool
pool::RelationalStorageService::ObjectReader::readPoolPolymorphicPointer( ObjectRelationalReadDataBuffer& dataBuffer,
                                                                          void* objectData,
                                                                          int oid,
                                                                          bool anyObject,
                                                                          pool::DataCallBack& callback ){
  const std::string& className = (*dataBuffer.parentMetadata())[dataBuffer.parentMetadataOffset()];
  if( className == RelationalStorageServiceNames::polyPtrNullLabel() ) return true;
  Reflex::Type ptrType = Reflex::Type::ByName(className);
  pool::RelationalStorageService::ObjectRelationalReadDataBuffer* newDataBuffer = 0;
  if( ptrType ){
    const pool::ObjectRelationalMappingElement& classMapping =  m_mappingManager.mappingForDependingClass( ptrType, m_dataBuffer->dictionary() );
    newDataBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_context);
    dataBuffer.dependentObjectBuffers().push_back( newDataBuffer );
    newDataBuffer->metadata() = dataBuffer.metadata();
    newDataBuffer->build( m_schema, classMapping, ptrType );
    ROOT::Reflex::Type poolPtrType = dataBuffer.dictionary();
    ROOT::Reflex::Object poolPtrObj(poolPtrType,objectData);
    poolPtrObj.Invoke("reset");

    ROOT::Reflex::Member ptrMember = poolPtrType.MemberByName("m_loader");
    RelationalPtrLoader* loader = new RelationalPtrLoader(*this,*newDataBuffer,objectData,oid,anyObject,callback);
    *(void**)(static_cast<char*>(objectData)+ptrMember.Offset())=loader;
    m_poolPointerLoaders.insert(loader);
    return true;
  }

  return false;
}

coral::IQuery*
pool::RelationalStorageService::ObjectReader::makeQuery( ObjectRelationalReadDataBuffer& dataBuffer,
                                                         int oid,
                                                         bool anyObject){
  coral::IQuery* query = dataBuffer.table().newQuery();
  const coral::AttributeListSpecification& outputSpecification = dataBuffer.dataSpecification();
  for ( coral::AttributeListSpecification::const_iterator iSpec = outputSpecification.begin();
        iSpec != outputSpecification.end(); ++iSpec ) {
    query->addToOutputList( iSpec->name() );
    query->defineOutputType( iSpec->name(),iSpec->typeName());
  }

  coral::AttributeList& oidData = dataBuffer.whereDataForId();
  std::stringstream whereClause;
  bool more = false;
  for(coral::AttributeList::iterator iAttr=oidData.begin();
      iAttr!=oidData.end();++iAttr){
    if(more) whereClause << " AND ";
    whereClause << iAttr->specification().name();
    if(iAttr->specification().typeName()==coral::AttributeSpecification::typeNameForId( typeid(std::string) )){
      iAttr->data<std::string>() = dataBuffer.metadata()[0];
      whereClause << "=";
    } else {
      iAttr->data< int >() = oid;
      if ( anyObject ) {
        whereClause << ">";
        query->addToOrderList( iAttr->specification().name() );
      }
      else whereClause << "=";
    }
    whereClause << " :" << iAttr->specification().name();
    more = true;
  }
  if ( anyObject ) {
    query->limitReturnedRows( 1, 0 );
  }  
  query->setCondition( whereClause.str(), oidData );
  return query;
}

bool
pool::RelationalStorageService::ObjectReader::readObjectData( coral::ICursor& queryData,
                                                              ObjectRelationalReadDataBuffer& dataBuffer,
                                                              void* objectData,
                                                              int& oid,
                                                              pool::DataCallBack& callback ){
  const coral::AttributeList& row = queryData.currentRow();

  // Get the OID
  coral::AttributeList::const_iterator iAttribute = row.begin();
  if(dataBuffer.mapping().isDependentTree()){
    ++iAttribute; // in this case, ignoring the first column (ref_column)  
  }
  oid = iAttribute->data< int >();
  
  ++iAttribute;

  // Copy the data
  std::vector<std::pair<int,int> >::const_iterator iOffset = dataBuffer.offsets().begin();
  while ( iAttribute != row.end() ) {
    int offset = iOffset->first;
    if ( offset < 0 ) {
      pool::Token::OID_t toid;
      toid.first = iAttribute->data<int>();
      ++iAttribute;
      ++iOffset;
      toid.second = iAttribute->data<int>();
      void* refAddress = static_cast< char* >( objectData ) - ( 1 + offset );
      // Call here the ref handler to get the oid...
      pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
      callback.notify( pool::DataCallBack::GET, pool::DataCallBack::REFERENCE, m_refDict, refAddress, &toid );
    } else {
      if( iOffset->second < 0 ) {
        if( iAttribute->specification().type() == typeid(coral::Blob) ){
          dataBuffer.readBlob( offset, iAttribute->data<coral::Blob>(), static_cast< char* >( objectData ) + offset );
        } else {
          iAttribute->copyValueToAddress( static_cast< char* >( objectData ) + offset );
        }
      } else {
        dataBuffer.metadata()[ iOffset->second ] = iAttribute->data<std::string>(); 
      }
      
    }
    ++iAttribute;
    ++iOffset;
  }
  return true;
}

bool
pool::RelationalStorageService::ObjectReader::processDependentBuffers( ObjectRelationalReadDataBuffer& dataBuffer,
                                                                       void* objectData,
                                                                       int oid,
                                                                       bool anyObject,
                                                                       pool::DataCallBack& callback ){
  // Loop over the arrays now...
  std::vector< int > identity( 1, oid );
  for ( pool::RelationalStorageService::ObjectRelationalReadDataBuffer::iterator iBuffer = dataBuffer.begin();
        iBuffer != dataBuffer.end(); ++iBuffer ) {
    (*iBuffer)->containerHandler().clear(static_cast< char* >( objectData ) + (*iBuffer)->offsetToParent());
    if ( ! this->readArray( identity, **iBuffer,
                            static_cast< char* >( objectData ) + (*iBuffer)->offsetToParent(),
                            callback ) ) return false;
  }
  // Loop over the pool pointers now...
  for ( std::vector<ObjectRelationalReadDataBuffer*>::const_iterator iBuffer = dataBuffer.poolPointerBuffers().begin();
        iBuffer != dataBuffer.poolPointerBuffers().end(); ++iBuffer ) {
    if ( ! this->readPoolPointer( *(*iBuffer),
                                  static_cast< char* >( objectData ) + (*iBuffer)->offsetToParent(),
                                  oid,
                                  anyObject,
                                  callback ) ) return false;
  }
  // Loop over the pool polymorphic pointers now...
  for ( std::vector<ObjectRelationalReadDataBuffer*>::const_iterator iBuffer = dataBuffer.poolPolymorphicPointerBuffers().begin();
        iBuffer != dataBuffer.poolPolymorphicPointerBuffers().end(); ++iBuffer ) {
    if ( ! this->readPoolPolymorphicPointer( *(*iBuffer),
                                             static_cast< char* >( objectData ) + (*iBuffer)->offsetToParent(),
                                             oid,
                                             anyObject,
                                             callback ) ) return false;
  }
  return true;
}

