#ifndef OBJECTRELATIONALDATABUFFER_H
#define OBJECTRELATIONALDATABUFFER_H

#include <vector>
#include <map>
#include "Reflex/Type.h"
#include "RVersion.h"

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
 typedef ROOT::Reflex::Type TypeH;
using namespace ROOT;
#else
 typedef Reflex::Type TypeH;
#endif

namespace coral {

  class ITable;
  class ISchema;
  class AttributeListSpecification;
  class AttributeList;
  class Blob;
}

namespace pool {

  class ObjectRelationalMappingElement;
  class IBlobWriter;
  class Context;

  namespace RelationalStorageService {

    class IContainerHandler;

    class ObjectRelationalDataBuffer {
    public:
      /// Constuctor
      explicit ObjectRelationalDataBuffer( pool::Context* context );

      /// Destructor
      ~ObjectRelationalDataBuffer();

      /// Builds the structure given a mapping element and the dictionary information
      bool build( coral::ISchema& schema,
                  const pool::ObjectRelationalMappingElement& mapping,
                  const TypeH& dictionary );

      /// Returns the data buffer
      coral::AttributeList& data();

      /// Returns the offset vector
      const std::vector<std::pair<int, int> >& offsets() const;

      /// Returns the metadata vector
      const std::vector< std::string >& metadata() const;
      
      /// Returns the metadata vector
      std::vector< std::string >& metadata();

      /// Returns the number of items belonging to the key (in case of an associative container)
      int keyItems() const;

      /// Returns the offset w.r.t. the parent object
      int offsetToParent() const;

      /// Returns the associoted table handle
      coral::ITable& table();

      /// Returns the associated dictionary information
      const TypeH& dictionary() const;

      /// Returns the container handler
      IContainerHandler& containerHandler();

      /// Returns a flag indicating whether the starting address is the key of an associative container
      bool isKeyData() const;

      /// Returns a blob version of the specified data, using the blob streamer registered in the pool context 
      const coral::Blob& writeBlob(int offset, const void* data);

      const std::string& updateClause() const;

      const std::string& idClause() const;

    public:
      typedef std::vector< ObjectRelationalDataBuffer* >::iterator iterator;
      iterator begin();
      iterator end();
      size_t size() const;
      const std::vector< ObjectRelationalDataBuffer* >& poolPointerBuffers() const;
      const std::vector< ObjectRelationalDataBuffer* >& poolPolymorphicPointerBuffers() const;

      const ObjectRelationalMappingElement& mapping() const;

    private:
      /// builds a buffer entry
      int buildBuffer( const pool::ObjectRelationalMappingElement& mapping,
                       const TypeH& dictionary,
                       coral::ISchema& schema,
                       int offset,
                       bool keyData );

      /// builds an array buffer
      bool buildArrayBuffer( const pool::ObjectRelationalMappingElement& mapping,
                             const TypeH& dictionary,
                             coral::ISchema& schema,
                             int offset,
                             bool keyData );
      
      /// builds a C-array buffer
      bool buildCArrayBuffer( const pool::ObjectRelationalMappingElement& mapping,
                             const TypeH& dictionary,
                             coral::ISchema& schema,
                             int offset,
                             bool keyData );
      /// builds a Pool Ptr buffer
      bool buildPoolPointerBuffer( const pool::ObjectRelationalMappingElement& mapping,
                                   const ROOT::Reflex::Type& dictionary,
                                   coral::ISchema& schema,
                                   int offset,
                                   bool keyData );

      /// builds a Pool PolyPtr buffer
      bool buildPoolPolymorphicPointerBuffer( const pool::ObjectRelationalMappingElement& topLevelMapping,
                                              const pool::ObjectRelationalMappingElement& polyPtrMapping,
                                              const ROOT::Reflex::Type& dictionary,
                                              int offset );

      /// add the column to the buffered update clause
      void appendToUpdateClause(const std::string& columnName);

      static bool existAttribute(const std::string& attributeName, const coral::AttributeList& data);

      private:
      
      pool::Context* m_context;
      const pool::ObjectRelationalMappingElement* m_mapping;
      coral::AttributeListSpecification* m_spec;
      std::vector< std::pair<int,int> > m_offsets;
      std::vector< std::string > m_metadata;
      coral::AttributeList* m_data;
      coral::ITable* m_table;
      TypeH m_dictionary;
      int m_offsetToParent;
      int m_keyItems;
      bool m_keyData;
      IContainerHandler* m_containerHandler;
      std::vector< ObjectRelationalDataBuffer* > m_arrayBuffers;
      std::map< int, pool::IBlobWriter* > m_blobWriters;
      std::vector< ObjectRelationalDataBuffer* > m_poolPointerBuffers;
      std::vector< ObjectRelationalDataBuffer* > m_poolPolymorphicPointerBuffers;
     std::string m_updateClause;
      std::string m_idClause;
    };

  }

}

#endif
