#ifndef RELATIONALPOOLCONTAINER_H
#define RELATIONALPOOLCONTAINER_H

#include "StorageSvc/DbContainerImp.h"

namespace pool {

  namespace RelationalStorageService {

    class PoolBaseContainerManager;
    class PoolDatabaseManager;

    /** @class RelationalPoolContainer RelationalPoolContainer.h
     *
     * Extension of the default implementation DbContainerImp for the object storage
     * using the POOL Relational backend
     */

    class RelationalPoolContainer : public pool::DbContainerImp
    {
    public:
      /// Constructor
      explicit RelationalPoolContainer( pool::IOODatabase* idb );

      /// Destructor
      virtual ~RelationalPoolContainer();

      /// Ask if a given shape is supported
      pool::DbStatus isShapeSupported( const pool::DbTypeInfo* type ) const;

      /// Open the container
      pool::DbStatus open( const pool::DbDatabase& dbH,
                           const std::string& name,
                           const pool::DbTypeInfo* info,
                           pool::DbAccessMode mode );

      /// Closes the container
      pool::DbStatus close();

      /// Define selection criteria
      pool::DbStatus select( pool::DbSelect& criteria );

      /// Number of entries within the container
      long long size();
      
      /// Number of record in the container
      long long nextRecordId();

      /// Fetch next object address of the selection to set token
      virtual pool::DbStatus fetch( pool::DbSelect& selection );

      /// Fetch a column identified by its link in the container
      virtual pool::DbStatus fetch( const pool::DbLink& linkH,
                                    pool::DbLink& stmt );
      
      /** Find object within the container and load it into memory
       *
       * @param  call     [IN]   Callback to load data
       * @param  linkH     [IN]   Preferred object OID
       * @param  oid      [OUT]   Actual object OID
       * @param  mode      [IN]   Object access mode
       * @param  any_next  [IN]   On selection, objects may be skipped.
       *                          If objects are skipped, the actual oid
       *                          will differ from the preferred oid.
       * @return Status code indicating success or failure.
       */
      pool::DbStatus load( pool::DataCallBack* call,
                           const pool::DbLink& lnkH, 
                           pool::DbLink&       oid,
                           pool::DbAccessMode  mod,
                           bool any_next );
 
      /// Find object by object identifier and load it into memory
      /** @param  call      [IN]   Callback to load data
       * @param  oid      [OUT]   Object OID
       * @param  mode      [IN]   Object access mode
       *
       * @return Status code indicating success or failure.
       */
      pool::DbStatus loadObject( pool::DataCallBack* call,
                                 pool::DbLink& oid,
                                 pool::DbAccessMode mode);
      
      /// Replace the entire content of the container with data mapped with the specified class info pairs
      /** @param classInfo      [IN]   pairs first=className second=mapping version
       *
       * @return Status code indicating success or failure.
       */
      pool::DbStatus updateContent(const std::vector<std::pair<std::string,std::string> >& classInfo,
                                                                                      bool recreate, 
                                                                                      bool dry=false);

      /// Execute end of object modification requests during a transaction
      /** @param refTr    [IN]  Transaction reference
       *
       * @return DbStatus code indicating success or failure.
       */
      pool::DbStatus endTransaction( pool::DbTransaction& refTr );

      std::string mappingVersion();

      std::string className();

      protected:
      /// Commit single entry to container
      pool::DbStatus writeObject( pool::DbContainerImp::TransactionStack::value_type& entry );

      /// Destroy persistent object in the container
      pool::DbStatus destroyObject( pool::DbContainerImp::TransactionStack::value_type& entry );

      /// Updates an existing object in the database
      pool::DbStatus updateObject( pool::DbContainerImp::TransactionStack::value_type& entry );

      // Returns the container manager
      PoolBaseContainerManager& containerManager();

    protected:
      /// Loads the correct container from the database manager
      virtual PoolBaseContainerManager* createContainerManager( PoolDatabaseManager& databaseManager,
                                                                const std::string& containerName,
                                                                const pool::DbTypeInfo* info ) = 0;

    private:
      /// Loads an object
      pool::DbStatus loadAnObject( pool::DataCallBack& call,
                                   int& oid,
                                   bool any_next );

      /// Writes or updates an object
      pool::DbStatus writeOrUpdateObject( pool::DbContainerImp::TransactionStack::value_type& entry,
                                          bool write );

    private:
      /// The read-only flag
      bool m_readOnly;

      /// The pointer to the database manager in use
      PoolDatabaseManager* m_databaseManager;

      /// The pointer to the parent database
      pool::DbDatabase* m_databaseContext;
    };

  }

}

#endif
