// $Id: DbCallBack.h,v 1.15 2007/11/20 16:43:06 frankb Exp $
//====================================================================
//  DbCallBack call back class definitions
//--------------------------------------------------------------------
//
//  Package    : StorageSvc  (The POOL project)
//  @author      M.Frank
//====================================================================
#ifndef POOL_DBCALLBACK_H
#define POOL_DBCALLBACK_H 1

// Framework include files
#include "StorageSvc/DbStatus.h"
#include "StorageSvc/DbObject.h"
#include "StorageSvc/DataCallBack.h"

/* 
 * POOL namespace declaration
 */
namespace pool  {

  // Forward declarations
  class DataHandler;
  class DbContainer;
  class DbTypeInfo;

  struct DbObjectEnv    {
    const DbContainer*              cntH;
    const DbObjectHandle<DbObject>* objH;
    const DbTypeInfo*               typH;
    /// Standard Destructor
    DbObjectEnv(const DbContainer* pCntH, 
                const DbObjectHandle<DbObject>* pObjH,
                const DbTypeInfo* pType)
    : cntH(pCntH), objH(pObjH), typH(pType) {}
    /// Standard Destructor
    DbObjectEnv(const DbObjectEnv& c) 
    : cntH(c.cntH), objH(c.objH), typH(c.typH) {}
  };

  /** @class DbCallBack DbCallBack.h StorageSvc/DbCallBack.h
    *
    * Description of class DbCallBack:
    * Definition of the class, which is able to read objects from the 
    * Database on an "as is" bases, i.e. in the format they were written.
    * This implicitly assumes, that the transient corresponds to the
    * persistent object mapping and that objects are sub-classes of the
    * DbObject class.
    *
    * @author  M.Frank
    * @version 1.0
    */
  class DbCallBack : public DataCallBack  {
  private:
    /// Object environment
    DbObjectEnv   m_env;
    /// Object pointer
    DbObject*     m_pObject;
    /// Data load/save handler
    DataHandler*  m_dataHandler;
  public:

    /// Constructor with initializing arguments
    DbCallBack(const DbContainer*              pCntH,
               const DbObjectHandle<DbObject>* pObjH,
               const DbTypeInfo*               pType);
    /// Copy Destructor
    DbCallBack(const DbCallBack& copy);
    /// Standard Destructor
    virtual ~DbCallBack();
    /// Returns the type of the transient object
    TypeH transientType() const;
    /// Abstract copy
    virtual DataCallBack* clone() const       { return new DbCallBack(*this);}
    /// Pass object for conversion data to the callback
    /**    @param ptr            [IN] Reference to object
    */
    virtual void setObject(const void* ptr)     { m_pObject = (DbObject*)ptr;}

    /// Pass shape object for conversion data to the callback
    /**    @param shape          [IN] Reference to shape object
    */
    virtual void setShape(ShapeH shape)         
    { m_env.typH = (const DbTypeInfo*)shape;                                 }

    /// Retrieve shape object from the callback
    virtual ShapeH shape()  const               { return ShapeH(m_env.typH); }

    /// Pass reference to data service for optional reference resolution
    /**    @param pSvc           [IN] Reference to Data service object
    */
    virtual void setHandler(DataHandler* pSvc)  { m_dataHandler = pSvc;      }

    /// Access optional supplied data handler
    virtual DataHandler* handler() const        { return m_dataHandler;      }

    /// Retrieve object from the callback
    virtual const void* object()  const         { return m_pObject;          }

    /// Access current type information
    const DbTypeInfo*  currentTypH() const      { return m_env.typH;         }

    /// Access current container information
    const DbContainer* currentCntH() const      { return m_env.cntH;         }

    /// Access current object information
    const DbObjectHandle<DbObject>* currentObjH() const { return m_env.objH; }

    /// Callback when a read sequence should be started
    /** @param action_type    [IN] Action type: PUT/GET
      * @param data           [IN] Reference to persistent data location (on Write)
      * @param context        [IN] Data context for subsequent calls
      *
      * @return DbStatus indicating success or failure.
      */
    virtual DbStatus start(CallType action_type, void* data, void** context);

    /// Callback to retrieve the absolute address of a column
    /** @param action_type    [IN] Action type: PUT/GET
      * @param col_identifier [IN] Reference to persistent column to be processed.
      * @param col_number     [IN] Number of persistent column.
      * @param context        [IN] Data context returned by "start"
      * @param data          [OUT] Reference to data location
      *
      * @return DbStatus indicating success or failure.
      *         If SKIP is returned, the column will not be processed.
      */
    virtual DbStatus bind(CallType             action_type,    // GET or PUT
                          const DbColumn*      col_identifier, // Col ID
                          int                  col_number,     // Col Number
                          void*                context,        // Start context
                          void**               data);          // data pointer

    /// Callback when a read sequence should be started
    /** @param action    [IN] Action type: PUT/GET
      * @param context   [IN] Data context returned by "start"
      *
      * @return DbStatus indicating success or failure.
      */
    virtual DbStatus end(CallType action, void* context);

    /// Callback to handle on the fly object manipulation during save/load operations.
    /** @param action_type    [IN] Action type: PUT/GET
      * @param data_type      [IN] DataType of the object to be processed
      * @param reflection     [IN] Reflection class information if availible
      * @param object_ptr     [IN] Pointer to the transient object.
      * @param data       [IN/OUT] Optional data for processing the request.
      *
      * @return DbStatus indicating success or failure.
      */
    virtual DbStatus notify(CallType  action_type,    // GET or PUT
                            int       data_type,      // data type
                            TypeH     reflection,     // Reflection information
                            void*     object_ptr,     // object ptr
                            void*     data);          // data pointer
  };
}      // End namespace pool
#endif // POOL_DBCALLBACK_H
