// $Id: DbObjectAccessor.h,v 1.7 2007/11/20 16:43:06 frankb Exp $
//====================================================================
//
//  @author      M.Frank
//====================================================================
#ifndef POOL_DBOBJECTACCESSOR_H
#define POOL_DBOBJECTACCESSOR_H 1

#ifndef POOL_DBOBJECT_H
#error "The header file DbObjectAccessor can only be included by DbObject.h"
#endif

/*
 * POOL namespace declaration
 */
namespace pool  {

  // Forward declarations
  class Token;

  /** Db objects: class DbObjectAccessor

    Description:
    Access class to implement handle actions

    @author  M.Frank
    @version 1.0
  */
  class DbObjectAccessor   {
  public:
    /// Open object with handle or callback
    /** Open object using either handle or callback starting
      * from the session handle of the originating object.
      *
      * @param objH      [IN/OUT] Handle to the object to be loaded.
      * @param callH     [IN/OUT] Callback to the object to be loaded.
      * @param sessionH  [IN]     Handle to source object's session.
      * @param pfn       [IN]     Physical file name of the database.
      *                           (if length==0: token's database is used.)
      * @param tokenH    [IN]     Reference to token of the target object.
      * @param mod       [IN]     Desired opening mode.
      *
      * @return DbStatus code indicating success or failure.
      */
    static DbStatus open( DbObjectHandle<DbObject>* objH,
                          DataCallBack*             callH,
                          const DbSession&          sessionH,
                          const std::string&        pfn,
                          const Token&              tokenH, 
                          DbAccessMode              mod);

    /// Open object with handle or callback
    /** Open object using either handle or callback starting
      * from the domain handle of the originating object.
      *
      * @param objH      [IN/OUT] Handle to the object to be loaded.
      * @param callH     [IN/OUT] Callback to the object to be loaded.
      * @param domH      [IN]     Handle to source object's domain.
      * @param pfn       [IN]     Physical file name of the database.
      *                           (if length==0: token's database is used.)
      * @param tokenH    [IN]     Reference to token of the target object.
      * @param mod       [IN]     Desired opening mode.
      *
      * @return DbStatus code indicating success or failure.
      */
    static DbStatus open( DbObjectHandle<DbObject>* objH,
                          DataCallBack*             callH,
                          const DbDomain&           domH,
                          const std::string&        pfn,
                          const Token&              tokenH, 
                          DbAccessMode              mod);

    /// Open object with handle or callback
    /** Open object using either handle or callback starting
      * from the database handle of the originating object.
      *
      * @param objH      [IN/OUT] Handle to the object to be loaded.
      * @param callH     [IN/OUT] Callback to the object to be loaded.
      * @param dbH       [IN]     Handle to source object's database.
      * @param tokenH    [IN]     Reference to token of the target object.
      * @param mod       [IN]     Desired opening mode.
      *
      * @return DbStatus code indicating success or failure.
      */
    static DbStatus open( DbObjectHandle<DbObject>* objH,
                          DataCallBack*             callH,
                          const DbDatabase&         dbH,  
                          const Token&              tokenH, 
                          DbAccessMode              mod);

    /// Open object with handle or callback
    /** Open object using either handle or callback starting
      * from the container handle of the originating object.
      *
      * @param objH      [IN/OUT] Handle to the object to be loaded.
      * @param callH     [IN/OUT] Callback to the object to be loaded.
      * @param cntH      [IN]     Handle to source object's container.
      * @param tokenH    [IN]     Reference to token of the target object.
      * @param mod       [IN]     Desired opening mode.
      *
      * @return DbStatus code indicating success or failure.
      */
    static DbStatus open( DbObjectHandle<DbObject>* objH,
                          DataCallBack*             callH,
                          const DbContainer&        cntH,
                          const Token&              tokenH, 
                          DbAccessMode              mod);

    /// Open object with handle or callback
    /** Open object using either handle or callback starting
      * from the database handle of the originating object.
      *
      * @param objH      [IN/OUT] Handle to the object to be loaded.
      * @param callH     [IN/OUT] Callback to the object to be loaded.
      * @param dbH       [IN]     Handle to source object's database.
      * @param linkH     [IN]     Reference to link handle of target object.
      * @param mod       [IN]     Desired opening mode.
      *
      * @return DbStatus code indicating success or failure.
      */
    static DbStatus open( DbObjectHandle<DbObject>* objH,
                          DataCallBack*             callH,
                          const DbDatabase&         dbH,  
                          const DbLink&             linkH, 
                          DbAccessMode              mod);

    /// Open object with handle or callback
    /** Open object using either handle or callback starting
      * from the container handle of the originating object.
      *
      * @param objH      [IN/OUT] Handle to the object to be loaded.
      * @param callH     [IN/OUT] Callback to the object to be loaded.
      * @param cntH      [IN]     Handle to source object's container.
      * @param linkH     [IN]     Reference to link handle of target object.
      * @param mod       [IN]     Desired opening mode.
      *
      * @return DbStatus code indicating success or failure.
      */
    static DbStatus open( DbObjectHandle<DbObject>* objH,
                          DataCallBack*             callH,
                          const DbContainer&        cntH,
                          const DbLink&             linkH, 
                          DbAccessMode              mod);

    /// Helper for handles to keep up with reference counting
    static void switchPointer(    const DbObjectSwitch& cast);
    
    /// Retrieve hosting container
    static const DbContainer& containedIn(const DbObject* pObj);
    
    /// Access object oid
    static DbLink& objectOid(const DbObject* ptr);

    /// Add persistent association entry
    static DbStatus makeObjectLink( const DbObject*     pObj,
                                    const Token*        pToken,
                                    DbLink&             tokenH);
    
    /// Validate an object association
    static DbStatus getObjectLink(  const DbObject*     pObj,
                                    const DbLink&       tokenH,
                                    Token*              pToken);
  };
}       // End namespace pool
#endif  // POOL_DBOBJECTACCESSOR_H
