// $Id: IDbContainer.h,v 1.21 2007/11/28 16:19:32 frankb Exp $
//
//  Package    : pool/StorageSvc (The pool framework)
//
//  @author      M.Frank
//
//====================================================================
#ifndef POOL_IDBCONTAINER_H
#define POOL_IDBCONTAINER_H
// $Header: /cvs/PF/pool/StorageSvc/StorageSvc/IDbContainer.h,v 1.21 2007/11/28 16:19:32 frankb Exp $

// Framework include files
#include "StorageSvc/pool.h"
#include "StorageSvc/DbObject.h"

/*
 *   POOL namespace declaration
 */
namespace pool    {

  // Forward declarations
  class DbTypeInfo;
  class DbDatabase;
  class DbContainer;
  class DataCallBack;
  class DbTransaction;
  class IDbCursor;
  class DbOption;
  class DbSelect;

  /** @class IDbContainer IDbContainer.h StorageSvc/IDbContainer.h 
    *
    * Description:
    *
    * Interface to the implementation specific part of a container object
    * objects.
    *
    * @author  M.Frank
    * @version 1.0
    */
  class IDbContainer    {
  protected:
    /// Destructor (called only by sub-classes)
    virtual ~IDbContainer()   {     }

  public:
    /// Release the technology specific implementation
    virtual void release() = 0;
    /// Access to container size
    virtual long long int size() = 0;
    /// Ask if a given shape is supported
    virtual DbStatus isShapeSupported(const DbTypeInfo* typ) const = 0;
    /// Set options
    virtual DbStatus setOption(const DbOption& refOpt) = 0;
    /// Access options
    virtual DbStatus getOption(DbOption& refOpt) const = 0;
    /// In place allocation of raw memory
    virtual void* allocate( unsigned long siz, 
                            DbContainer&  cntH,
                            DataCallBack* call) = 0;
    /// In place allocation of object location
    virtual DbStatus allocate(DbContainer& cntH, 
                              DataCallBack* call,
                              DbLink& oid) = 0;
    /// In place deletion of raw memory
    virtual DbStatus free(  void* ptr,
                            DbContainer&  cntH,
                            DataCallBack* call) = 0;
    /// Close the container
    virtual DbStatus close() = 0;
    /// Open the container
    virtual DbStatus open(  const DbDatabase& dbH, 
                            const std::string& nam, 
                            const DbTypeInfo* info, 
                            DbAccessMode mode) = 0;
    /// Perform UPDATE select
    virtual DbStatus update(  DbSelect& sel) = 0;
    /// Perform DELETE select
    virtual DbStatus destroy( DbSelect& sel) = 0;
    /// Define selection criteria
    virtual DbStatus select(  DbSelect& criteria) = 0;
    /// Fetch next object address of the selection to set token
    virtual DbStatus fetch(DbSelect& sel) = 0;

    /// Find object within the container and load it into memory
    /** @param  call      [IN]   Callback to load data
      * @param  linkH     [IN]   Preferred object OID
      * @param  oid      [OUT]   Actual object OID
      * @param  mode      [IN]   Object access mode
      * @param  any_next  [IN]   On selection, objects may be skipped.
      *                          If objects are skipped, the actual oid
      *                          will differ from the preferred oid.
      * @return Status code indicating success or failure.
      */
    virtual DbStatus load(  DataCallBack* call, 
                            const DbLink& linkH,
                            DbLink&       oid,
                            DbAccessMode  mode,
                            bool          any_next=false) = 0;

    /// Save new object in the container and return its handle
    /** @param  cntH      [IN]   Handle to container object.
      * @param  call      [IN]   Data callback to feed data.
      * @param  linkH     [OUT]  Internal OID to identify object.
      *
      * @return DbStatus code indicating success or failure.
      */
    virtual DbStatus save(  DbContainer&  cntH,
                            DataCallBack* call,
                            DbLink& linkH) = 0;

    /// Add object to the container
    virtual DbStatus save(  const DbObjectHandle<DbObject>& objH) = 0;
    /// Update existing object in the container
    /** @param cntH      [IN]     Valid handle to container 
      * @param call      [IN]     Data transformation callback
      * @param objH      [IN]     Object handle
      *
      * @return Status code indicating success or failure.
      */
    virtual DbStatus update(DbContainer&  cntH,
                            DataCallBack* call,
                            const DbObjectHandle<DbObject>& objH) = 0;

    /// Update existing object in the container
    /** @param cntH      [IN]     Valid handle to container 
      * @param call      [IN]     Data transformation callback
      * @param linkH     [IN/OUT] Object identifier
      *
      * @return Status code indicating success or failure.
      */
    virtual DbStatus update(DbContainer&  cntH,
                            DataCallBack* call,
                            const DbLink& linkH) = 0;

    /// Destroy an object in a container
    /** @param linkH     [IN]     Object identifier
      *
      * @return Status code indicating success or failure.
      */
    virtual DbStatus destroy( const DbLink& linkH) = 0;

    /// Start/Commit/Rollback Database Transaction
    /** @param refTransaction [IN/OUT]  Transaction context.
      *
      * @return DbStatus code indicating success or failure.  
      */
    virtual DbStatus transAct(DbTransaction& refTransaction) = 0;

    /// Query the pending transaction stack
    virtual bool updatesPending() const = 0;

    /// Create iterator cursor to loop over container elements
    virtual IDbCursor* makeCursor(const DbTypeInfo* typ) = 0;
  };
}      // End namespace pool
#endif // POOL_IDBCONTAINER_H
