// $Id: DbDatabaseObj.h,v 1.26 2008/05/13 14:39:32 witoldp Exp $
//====================================================================
//  DbDatabaseObj object definition
//--------------------------------------------------------------------
//
//  Package    : System (The POOL project)
//
//  Description: Generic data persistency
//
//  @author      M.Frank
//====================================================================
#ifndef POOL_DBDATABASEOBJ_H
#define POOL_DBDATABASEOBJ_H 1

// We need the definition of the domain for defining the Database
#include "StorageSvc/DbAccessObj.h"
#include "StorageSvc/DbDomain.h"
#include "StorageSvc/DbDatabase.h"
#include "StorageSvc/DbContainer.h"
#include "POOLCore/Guid.h"

// STL include files
#include <map>
#include <string>

/*
 *  POOL namespace declaration
 */
namespace pool    {

  // Forward declarations
  class DbToken;

  /** @class DbDatabaseObj DbDatabaseObj.h src/DbDatabaseObj.h
    *
    * Description:
    *
    * Implementation independent part of a Database object
    * objects.
    *
    * @author  M.Frank
    * @version 1.0
    */
  class DbDatabaseObj : public  DbAccessObj<std::string, DbContainerObj > {
  private:
    /// Reflection class identifier
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
    typedef const ROOT::Reflex::Type&                      TypeH;
#else
    typedef const Reflex::Type&                      TypeH;
#endif
    /// Database parameter definition
    typedef std::pair<std::string, std::string>      Parameter;
    /// Database parameter container definition
    typedef std::vector< Parameter >                 Parameters;
    /// Collection of retired database containers
    typedef std::vector< DbContainerObj >            Containers;
    /// Parameter map definition
    typedef std::map<std::string, std::string>       ParamMap;
    /// Definition of map with link elements
    typedef std::map< Guid , DbToken* >              LinkMap;
    /// Definition of array with link elements
    typedef std::vector<DbToken*>                    LinkVector;
    /// Definition of map with shape elements
    typedef std::map< Guid , const DbTypeInfo* >     ShapeMap;
    /// Definition of map with class elements
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
    typedef std::map<ROOT::Reflex::Type,const DbTypeInfo*> ClassMap;
#else
    typedef std::map<Reflex::Type,const DbTypeInfo*> ClassMap;
#endif
    /// Handle to domain
    DbDomain                      m_dom;
    /// Handle to link container
    DbContainer                   m_links;
    /// Handle to shapes container
    DbContainer                   m_shapes;
    /// Handle to the parameter container
    DbContainer                   m_params;
    /// Technology dependent implementation block
    IDbDatabase*                  m_info;
    /// Map containing all links
    LinkMap                       m_linkMap;
    /// Random access array containing all links
    LinkVector                    m_linkVec;
    /// Map containing all links
    ShapeMap                      m_shapeMap;
    /// Map with all cached file properties
    ParamMap                      m_paramMap;
    /// Map with all cached mappings between known reflection classes
    ClassMap                      m_classMap;
    /// Collection of retired database containers
    Containers                    m_retiredConts;
    /// Internal string representation type
    const DbTypeInfo*             m_string_t;
    /// Physical Database login
    std::string                   m_logon;
    /// Flag to mark ongoing Transactions
    bool                          m_transactOpen;
    /// File age counter
    int                           m_fileAge;
    /// Token describing the object
    Token*                        m_token;

  private:
    /// Perform cleanup of internal data structures
    DbStatus cleanup();
  public:
    /// Standard constructor of a Database obejct
    DbDatabaseObj( const DbDomain& dom, 
                   const std::string& pfn,
                   const std::string& fid,
                   DbAccessMode mode = pool::READ);
    /// Standard Destructor
    virtual ~DbDatabaseObj();
    /// Access to the logon string
    const std::string& logon() const  {  return m_logon;        }
    /// Access the size of the database: May be undefined for some technologies
    long long int size()  const;
    /// Access to technology dependent implementation
    IDbDatabase* info() const         {  return m_info;         }
    /// Access to domain handle (CONST)
    DbDomain& domain()                {  return m_dom;          }
    /// Allow query if Transaction is active
    bool transactionActive() const    {  return m_transactOpen; }
    /// Access age value
    int  age()  const                 {  return m_fileAge;      }
    /// Access the token of the database object
    const Token* token() const        {  return m_token;        }

    /// Open Database object
    DbStatus open();
    /// Re-open database with changing access permissions
    DbStatus reopen(DbAccessMode mode);
    /// Close database object
    DbStatus close();
    /// End database access, but still leave database accessible
    DbStatus retire();
    /// Check for pending updates
    bool updatesPending() const;
    /// Start/Commit/Rollback Database Transaction
    DbStatus transAct(DbTransaction& refTransaction);
    /// Retrieve association link from link container
    DbStatus getLink(const DbLink& lnkH, Token* pToken);
    /// Add association link to link container
    DbStatus makeLink(const Token* pToken, DbLink& refLink);
    /// Retrieve persistent type information by class handle
    const DbTypeInfo* objectShape(TypeH classH);
    /// Retrieve persistent type information by shape identifier
    const DbTypeInfo* objectShape(const Guid& nam);
    /// Retrieve persistent type information by container
    const DbTypeInfo* contShape(const std::string& nam);
    /// Add persistent type to the Database
    DbStatus addShape (const DbTypeInfo* pType);
    /// Access local container token (if container exists)
    const Token* cntToken(const std::string& cntName);
    /// Allow access to all known containers
    DbStatus containers(std::vector<const Token*>& conts,
                      bool intern);
    /// Allow access to all known associations between containers
    DbStatus associations(std::vector<const Token*>& conts);
    /// Allow access to all known shapes used by the database
    DbStatus shapes(std::vector<const DbTypeInfo*>& shaps);
    /// Retrieve the number of user parameters
    int nParam() const;
    /// Add a persistent parameter to the file
    DbStatus addParam(const std::string& nam, const std::string& val);
    /// Retrieve existing parameter by name
    DbStatus param(const std::string& nam, std::string& val);
    /// Retrieve all parameters
    DbStatus params(std::vector< std::pair<std::string, std::string> >& vals);
    /// Set options
    DbStatus setOption(const DbOption& refOpt);
    /// Access options
    DbStatus getOption(DbOption& refOpt);
    /// Update database age
    void setAge(int value);
  };
}       // End namespace pool
#endif  // POOL_DBDATABASEOBJ_H
