// $Id: DbMemoryIO.cpp,v 1.16 2008/10/21 08:39:59 frankb Exp $
//====================================================================
//  DbMemoryIO class implementation
//--------------------------------------------------------------------
//
//  Package    : StorageSvc (The pool framework)
//
//  @author      M.Frank
//
//====================================================================

static bool s_inited = false;

#include "StorageSvc/DbPrint.h"
#include "StorageSvc/DbMemoryIO.h"
#include <sys/types.h>
#include <iostream>
#include <cerrno>
#include <cstring>
#include <fcntl.h>
using namespace std;
using namespace pool;

DbMemoryIO* DbMemoryIO::instance()  {
  static DbMemoryIO s_inst;
  return &s_inst;
}

DbMemoryIO::DbMemoryIO() {
  s_inited = true;
}

DbMemoryIO::~DbMemoryIO()  {
  s_inited = false;
  unlinkAll();
}

int DbMemoryIO::open(const string& url, int mode, unsigned int /*flag*/)  {
  if ( s_inited )  {
    bool read_access = (mode&0xFFF)==O_RDONLY || 
                       (mode&O_RDWR)==O_RDWR  || 
#ifdef _WIN32
                       (mode&_O_APPEND)==_O_APPEND;
#else 
                       (mode&O_APPEND)==O_APPEND;
#endif
    Pathes::iterator i = m_pathMap.find(url);
    if ( i == m_pathMap.end() && (mode&O_CREAT))  {
      Stat s;
      int fd = m_pathMap.size()+0xFEED;
      m_pathMap.insert(make_pair(url,fd));
      s.st_name = url;
      s.st_mode = kReadPermission+kWritePermission+kFileExists;
      s.st_length = 16*1024;
      s.st_fptr = s.st_buff = s.st_end = (char*)malloc(size_t(s.st_length));
      s.st_size = 0;
      s.st_open = true;
      m_fdMap.insert(make_pair(fd,s));
      return fd;
    }
    else if ( mode&O_CREAT )  {
      errno = EEXIST;
      return -1;
    }
    else if ( i != m_pathMap.end() && read_access )  {
      Files::iterator j = m_fdMap.find((*i).second);
      if ( j != m_fdMap.end() )  {
        Stat& b = (*j).second;
        b.st_fptr = b.st_buff;
        b.st_end  = b.st_buff + b.st_size;
        b.st_open = true;
        return (*i).second;
      }
    }
  }
  else  {
    DbPrint err("DbMemoryIO");
    err << DbPrintLvl::Error << "Attempt to open memory based file after unloading library!"
        << DbPrint::endmsg;
  }
  errno = ENOENT;
  return -1;
}

int DbMemoryIO::access(const string& url, int mode)  {
  if ( s_inited )  {
    Pathes::iterator i = m_pathMap.find(url);
    if ( i != m_pathMap.end() )  {
      Files::iterator fit = m_fdMap.find((*i).second);
      if ( fit != m_fdMap.end() )  {
        Stat& b = (*fit).second;
        if ( mode==kFileExists )  {
          return 0;
        }
        else if ( (mode&kReadPermission) == (b.st_mode&kReadPermission) )  {
          return 0;
        }
        else if ( (mode&kWritePermission) == (b.st_mode&kWritePermission) )  {
          return 0;
        }
      }
    }
  }
  else  {
    DbPrint err("DbMemoryIO");
    err << DbPrintLvl::Error << "Attempt to inquire memory based file after unloading library!" << DbPrint::endmsg;
  }
  return -1;
}

int DbMemoryIO::unlink(const string& url)  {
  if ( s_inited )  {
    Pathes::iterator i = m_pathMap.find(url);
    if ( i != m_pathMap.end() )  {
      Files::iterator j = m_fdMap.find((*i).second);
      if ( j != m_fdMap.end() )  {
        Stat& b = (*j).second;
        if ( !(b.st_mode&kWritePermission) )  {
          errno = EACCES;
          return -1;
        }
        ::free(b.st_buff);
        m_fdMap.erase(j);
      }
      m_pathMap.erase(i);
      return 0;
    }
  }
  else  {
    DbPrint err("DbMemoryIO");
    err << DbPrintLvl::Error << "Attempt to unlink memory based file after unloading library!" << DbPrint::endmsg;
  }
  errno = ENOENT;
  return -1;
}

int DbMemoryIO::unlinkAll()  {
  if ( s_inited )  {
    while( m_pathMap.begin() != m_pathMap.end() ) {
      unlink((*m_pathMap.begin()).first);
    }
  }
  return 0;
}

int DbMemoryIO::close(int fd)  {
  if ( s_inited )  {
    Files::iterator j = m_fdMap.find(fd);
    if ( j != m_fdMap.end() )  {
      Stat& b = (*j).second;
      b.st_fptr = b.st_buff;
      b.st_end  = b.st_buff + b.st_size;
      b.st_open = false;
      return 0;
    }
    errno = EBADF;
    return -1;
  }
  else  {
    DbPrint err("DbMemoryIO");
    err << DbPrintLvl::Error << "Attempt to close memory based file after unloading library!" << DbPrint::endmsg;
  }
  return 0;
}

long long int DbMemoryIO::read(int fd, void* buff, long long int siz)   {
  if ( s_inited )  {
    Files::iterator j = m_fdMap.find(fd);
    if ( j != m_fdMap.end() )  {
      Stat& b = (*j).second;
      if ( b.st_fptr+siz < b.st_buff+b.st_size )  {
        ::memcpy(buff, (const void*)b.st_fptr, size_t(siz));
        b.st_fptr += siz;
        return siz;
      }
      else if ( b.st_fptr < b.st_buff+b.st_size )  {
        ::memcpy(buff, (const void*)b.st_fptr, b.st_buff+b.st_size-b.st_fptr);
        b.st_fptr = b.st_buff+b.st_size;
        return siz;
      }
    }
  }
  else  {
    DbPrint err("DbMemoryIO");
    err << DbPrintLvl::Error << "Attempt to read memory based file after unloading library!" << DbPrint::endmsg;
  }
  errno = EBADF;
  return -1;
}

long long int DbMemoryIO::write(int fd, const void* buff, long long int siz)  {
  if ( s_inited )  {
    Files::iterator j = m_fdMap.find(fd);
    if ( j != m_fdMap.end() )  {
      Stat& b = (*j).second;
      if ( b.st_fptr+siz > b.st_buff+b.st_length )  {
        long long int offset = b.st_fptr-b.st_buff;
        b.st_length = b.st_length+b.st_length+siz;
        b.st_buff   = (char*)::realloc((void*)b.st_buff, size_t(b.st_length));
        b.st_fptr   = b.st_buff + offset;
      }
      ::memcpy(b.st_fptr, buff, size_t(siz));
      b.st_fptr += siz;
      if ( b.st_fptr > b.st_end )  {
        b.st_end  = b.st_fptr;
        b.st_size = b.st_fptr-b.st_buff;
      }
      return siz;
    }
  }
  else  {
    DbPrint err("DbMemoryIO");
    err << DbPrintLvl::Error << "Attempt to write to memory based file after unloading library!" << DbPrint::endmsg;
  }
  errno = EBADF;
  return -1;
}

long long int DbMemoryIO::seek(int fd, long long int offset, int origin)  {
  if ( s_inited )  {
    Files::iterator j = m_fdMap.find(fd);
    if ( j != m_fdMap.end() )  {
      Stat& b = (*j).second;
      switch(origin)  {
        case SEEK_SET: // Beginning of file. 
          b.st_fptr = b.st_buff;
          b.st_fptr += offset;
          break;
        case SEEK_CUR: // Current position of file pointer.
          b.st_fptr += offset;
          break;
        case SEEK_END:
          b.st_fptr = b.st_buff + b.st_size;
          b.st_fptr += offset;
        default:
          errno = EINVAL;
          return -1;
      }
      return b.st_fptr - b.st_buff;
    }
  }
  else  {
    DbPrint err("DbMemoryIO");
    err << DbPrintLvl::Error << "Attempt to seek memory based file after unloading library!" << DbPrint::endmsg;
  }
  errno = EBADF;
  return -1;
}

int DbMemoryIO::stat(int fd, IOStat* buf)  {
  if ( s_inited )  {
    Files::iterator j = m_fdMap.find(fd);
    if ( j != m_fdMap.end() )  {
      if ( buf )  {
        Stat& b = (*j).second;
        memset(buf,0,sizeof(*buf));
        buf->st_size  = off_t(b.st_size);
        buf->st_mode  = b.st_mode;
        buf->st_nlink = 1;
        buf->st_ino   = fd;
        return 0;
      }
      errno = EINVAL;
      return -1;
    }
  }
  else  {
    DbPrint err("DbMemoryIO");
    err << DbPrintLvl::Error << "Attempt to stat memory based file after unloading library!" << DbPrint::endmsg;
  }
  errno = EBADF;
  return -1;
}

int DbMemoryIO::stat(const string& url, IOStat* buf)  {
  if ( s_inited )  {
    Pathes::iterator i = m_pathMap.find(url);
    if ( i != m_pathMap.end() )  {
      return stat((*i).second, buf);
    }
  }
  else  {
    DbPrint err("DbMemoryIO");
    err << DbPrintLvl::Error << "Attempt to stat memory based file after unloading library!" << DbPrint::endmsg;
  }
  errno = EBADF;
  return -1;
}


int DbMemoryIO::dump(const string& url)  {
  if ( s_inited )  {
    Pathes::iterator i = m_pathMap.find(url);
    if ( i != m_pathMap.end() )  {
      return dump((*i).second);
    }
  }
  else  {
    DbPrint err("DbMemoryIO");
    err << DbPrintLvl::Error << "Attempt to dump memory based file after unloading library!" << DbPrint::endmsg;
  }
  errno = EBADF;
  return -1;
}

int DbMemoryIO::dump(int fd)  {
  if ( s_inited )  {
    Files::iterator j = m_fdMap.find(fd);
    if ( j != m_fdMap.end() )  {
      Stat& b = (*j).second;
      DbPrint log("DbMemoryIO");
      log << DbPrintLvl::Always 
          << "File:   " << b.st_name        << DbPrint::endmsg
          << "Open:   " << b.st_open        << DbPrint::endmsg
          << "Mode:   " << b.st_mode        << DbPrint::endmsg
          << "Length: " << b.st_length      << DbPrint::endmsg
          << "Size:   " << b.st_size        << DbPrint::endmsg
          << "FPtr:   " << (void*)b.st_fptr << DbPrint::endmsg
          << "Buffer: " << (void*)b.st_buff << DbPrint::endmsg;
      return 0;
    }
  }
  return -1;
}

/// Access file information
DbMemoryIO::Stat* DbMemoryIO::info(int fd)   {
  if ( s_inited )  {
    Files::iterator j = m_fdMap.find(fd);
    if ( j != m_fdMap.end() )  {
      return &(*j).second;
    }
  }
  return 0;
}

/// Access file information
DbMemoryIO::Stat* DbMemoryIO::info(const string& url)  {
  if ( s_inited )  {
    Pathes::iterator i = m_pathMap.find(url);
    if ( i != m_pathMap.end() )  {
      return info((*i).second);
    }
  }
  return 0;
}
